
# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  pyomo-based dispatch strategy
"""

import os
import sys
import time as time_mod
import platform
from itertools import compress
import pprint

import numpy as np
import pyomo.environ as pyo
from pyomo.opt import SolverStatus, TerminationCondition

from ravenframework.utils import InputData, InputTypes

# allows pyomo to solve on threaded processes
import pyutilib.subprocess.GlobalData
pyutilib.subprocess.GlobalData.DEFINE_SIGNAL_HANDLERS_DEFAULT = False
from pyomo.common.errors import ApplicationError

from .Dispatcher import Dispatcher
from .DispatchState import DispatchState, NumpyState
try:
  import _utils as hutils
except (ModuleNotFoundError, ImportError):
  sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
  import _utils as hutils

# Choose solver; CBC is a great choice unless we're on Windows
if platform.system() == 'Windows':
  SOLVERS = ['glpk', 'cbc', 'ipopt']
else:
  SOLVERS = ['cbc', 'glpk', 'ipopt']


class Pyomo(Dispatcher):
  """
    Dispatches using rolling windows in Pyomo
  """
  naming_template = {'comp prod': '{comp}|{res}|prod',
                     'comp transfer': '{comp}|{res}|trans',
                     'comp max': '{comp}|{res}|max',
                     'comp ramp up': '{comp}|{res}|rampup',
                     'comp ramp down': '{comp}|{res}|rampdown',
                     'conservation': '{res}|consv',
                    }

  ### INITIALIZATION
  @classmethod
  def get_input_specs(cls):
    """
      Set acceptable input specifications.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = InputData.parameterInputFactory('pyomo', ordered=False, baseNode=None,
        descr=r"""The \texttt{pyomo} dispatcher uses analytic modeling and rolling windows to
        solve dispatch optimization with perfect information via the pyomo optimization library.""")
    specs.addSub(InputData.parameterInputFactory('rolling_window_length', contentType=InputTypes.IntegerType,
        descr=r"""Sets the length of the rolling window that the Pyomo optimization algorithm
        uses to break down histories. Longer window lengths will minimize boundary effects, such as
        nonoptimal storage dispatch, at the cost of slower optimization solves.
        Note that if the rolling window results in a window of length 1 (such as at the end of a history),
        this can cause problems for pyomo.
        \default{24}"""))
    specs.addSub(InputData.parameterInputFactory('debug_mode', contentType=InputTypes.BoolType,
        descr=r"""Enables additional printing in the pyomo dispatcher. Highly discouraged for production runs.
        \default{False}."""))
    specs.addSub(InputData.parameterInputFactory('solver', contentType=InputTypes.StringType,
        descr=r"""Indicates which solver should be used by pyomo. Options depend on individual installation.
        \default{'glpk' for Windows, 'cbc' otherwise}."""))
    # TODO specific for pyomo dispatcher
    return specs

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    self.name = 'PyomoDispatcher' # identifying name
    self.debug_mode = False       # whether to print additional information
    self._window_len = 24         # time window length to dispatch at a time # FIXME user input
    self._solver = None           # overwrite option for solver
    self._picard_limit = 10       # iterative solve limit

  def read_input(self, specs):
    """
      Read in input specifications.
      @ In, specs, RAVEN InputData, specifications
      @ Out, None
    """
    super().read_input(specs)

    window_len_node = specs.findFirst('rolling_window_length')
    if window_len_node is not None:
      self._window_len = window_len_node.value

    debug_node = specs.findFirst('debug_mode')
    if debug_node is not None:
      self.debug_mode = debug_node.value

    solver_node = specs.findFirst('solver')
    if solver_node is not None:
      self._solver = solver_node.value

    if self._solver is None:
      solvers_to_check = SOLVERS
    else:
      solvers_to_check = [self._solver]
    # check solver exists
    for solver in solvers_to_check:
      self._solver = solver
      found_solver = True
      try:
        if not pyo.SolverFactory(self._solver).available():
          found_solver = False
        else:
          break
      except ApplicationError:
        found_solver = False
    # NOTE: we probably need a consistent way to test and check viable solvers,
    # maybe through a unit test that mimics the model setup here. For now, I assume
    # that anything that shows as not available or starts with an underscore is not
    # viable, and it will crash if the solver can't solve our kinds of models.
    # This should only come up if the user is specifically requesting a solver, though,
    # the default glpk and cbc are tested.
    if not found_solver:
      all_options = pyo.SolverFactory._cls.keys() # TODO shorten to list of tested options?
      solver_filter = []
      for op in all_options:
        if op.startswith('_'): # These don't seem like legitimate options, based on testing
          solver_filter.append(False)
          continue
        try:
          solver_filter.append(pyo.SolverFactory(op).available())
        except (ApplicationError, NameError, ImportError):
          solver_filter.append(False)
      available = list(compress(all_options, solver_filter))
      msg = f'Requested solver "{self._solver}" was not found for pyomo dispatcher!'
      msg += f' Options MAY include: {available}'
      raise RuntimeError(msg)


  ### API
  def dispatch(self, case, components, sources, meta):
    """
      Performs dispatch.
      @ In, case, HERON Case, Case that this dispatch is part of
      @ In, components, list, HERON components available to the dispatch
      @ In, sources, list, HERON source (placeholders) for signals
      @ In, meta, dict, additional variables passed through
      @ Out, disp, DispatchScenario, resulting dispatch
    """
    t_start, t_end, t_num = self.get_time_discr()
    time = np.linspace(t_start, t_end, t_num) # Note we don't care about segment/cluster here
    resources = sorted(list(hutils.get_all_resources(components))) # list of all active resources
    # pre-build results structure
    ## we can use NumpyState here so we don't need to worry about a Pyomo model object
    dispatch = NumpyState()# dict((comp.name, dict((res, np.zeros(len(time))) for res in comp.get_resources())) for comp in components)
    dispatch.initialize(components, meta['HERON']['resource_indexer'], time)
    # rolling window
    start_index = 0
    final_index = len(time)
    # TODO window overlap!  ( )[ ] -> (   [  )   ]
    while start_index < final_index:
      end_index = start_index + self._window_len
      if end_index > final_index:
        end_index = final_index
      if end_index - start_index == 1:
        # TODO custom error raise for catching in DispatchManager?
        raise IOError("A rolling window of length 1 was requested, but this causes crashes in pyomo. " +
                      "Change the length of the rolling window to avoid length 1 histories.")
      specific_time = time[start_index:end_index]
      print('DEBUGG starting window {} to {}'.format(start_index, end_index))
      start = time_mod.time()
      # set initial storage levels
      initial_levels = {}
      for comp in components:
        if comp.get_interaction().is_type('Storage'):
          if start_index == 0:
            initial_levels[comp] = comp.get_interaction().get_initial_level(meta)
          else:
            initial_levels[comp] = subdisp[comp.name]['level'][comp.get_interaction().get_resource()][-1]
      # allow for converging solution iteratively
      converged = False
      conv_counter = 0
      previous = None
      while not converged:
        conv_counter += 1
        if conv_counter > self._picard_limit:
          break
        # dispatch
        subdisp = self.dispatch_window(specific_time, start_index,
                                      case, components, sources, resources,
                                      initial_levels, meta)
        # do we need a convergence criteria? Check now.
        if self.needs_convergence(components):
          print(f'DEBUGG iteratively solving window, iteration {conv_counter}/{self._picard_limit} ...')
          converged = self.check_converged(subdisp, previous, components)
          previous = subdisp
        else:
          converged = True

      end = time_mod.time()
      print('DEBUGG solve time: {} s'.format(end-start))
      # store result in corresponding part of dispatch
      for comp in components:
        for tag in comp.get_tracking_vars():
          for res, values in subdisp[comp.name][tag].items():
            dispatch.set_activity_vector(comp, res, values, tracker=tag, start_idx=start_index, end_idx=end_index)
      start_index = end_index
    return dispatch

  ### INTERNAL
  def dispatch_window(self, time, time_offset,
                      case, components, sources, resources,
                      initial_storage, meta):
    """
      Dispatches one part of a rolling window.
      @ In, time, np.array, value of time to evaluate
      @ In, time_offset, int, offset of the time index in the greater history
      @ In, case, HERON Case, Case that this dispatch is part of
      @ In, components, list, HERON components available to the dispatch
      @ In, sources, list, HERON source (placeholders) for signals
      @ In, resources, list, sorted list of all resources in problem
      @ In, initial_storage, dict, initial storage levels if any
      @ In, meta, dict, additional variables passed through
      @ Out, result, dict, results of window dispatch
    """
    # build the Pyomo model
    # TODO abstract this model as much as possible BEFORE, then concrete initialization per window
    m = pyo.ConcreteModel()
    # indices
    C = np.arange(0, len(components), dtype=int) # indexes component
    R = np.arange(0, len(resources), dtype=int) # indexes resources
    # T = np.arange(start_index, end_index, dtype=int) # indexes resources
    T = np.arange(0, len(time), dtype=int) # indexes resources
    m.C = pyo.Set(initialize=C)
    m.R = pyo.Set(initialize=R)
    m.T = pyo.Set(initialize=T)
    m.Times = time
    dt = m.Times[1] - m.Times[0] # TODO assumes consistent step sizing
    m.time_offset = time_offset
    m.resource_index_map = meta['HERON']['resource_indexer'] # maps the resource to its index WITHIN APPLICABLE components (sparse matrix)
                                                             #   e.g. component: {resource: local index}, ... etc}
    # properties
    m.Case = case
    m.Components = components
    m.Activity = PyomoState()
    m.Activity.initialize(m.Components, m.resource_index_map, m.Times, m)
    # constraints and variables
    for comp in components:
      # components using a governing strategy (not opt) are Parameters, not Variables
      # TODO should this come BEFORE or AFTER each dispatch opt solve?
      # -> responsive or proactive?
      intr = comp.get_interaction()
      if intr.is_governed():
        meta['request'] = {'component': comp, 'time': time}
        activity = intr.get_strategy().evaluate(meta)[0]['level']
        if intr.is_type('Storage'):
          self._create_production_param(m, comp, activity, tag='level')
          # set up "activity" rates (change in level over time, plus efficiency)
          rte2 = comp.get_sqrt_RTE() # square root of the round-trip efficiency
          L  = len(activity)
          deltas = np.zeros(L)
          deltas[1:] = activity[1:] - activity[:-1]
          deltas[0] = activity[0] - intr.get_initial_level(meta)
          # rate of charge
          # change sign, since increasing level means absorbing energy from system
          # also scale by RTE, since to get level increase you have to over-absorb
          charge = np.zeros(L)
          charge_mask = np.where(deltas > 0)
          charge[charge_mask] = - deltas[charge_mask] / dt / rte2
          self._create_production_param(m, comp, charge, tag='charge')
          # rate of discharge
          # change sign, since decreasing level means emitting energy into system
          # also scale by RTE, since level decrease yields less to system
          discharge = np.zeros(L)
          discharge_mask = np.where(deltas < 0)
          discharge[discharge_mask] = - deltas[discharge_mask] / dt * rte2
          self._create_production_param(m, comp, discharge, tag='discharge')
        else:
          self._create_production_param(m, comp, activity)
        continue
      # NOTE: "fixed" components could hypothetically be treated differently
      ## however, in order for the "production" variable for components to be treatable as the
      ## same as other production variables, we create components with limitation
      ## lowerbound == upperbound == capacity (just for "fixed" dispatch components)
      if intr.is_type('Storage'):
        self._create_storage(m, comp, initial_storage, meta)
      else:
        self._create_production(m, comp, meta) # variables
    self._create_conservation(m, resources, initial_storage, meta) # conservation of resources (e.g. production == consumption)
    self._create_objective(meta, m) # objective
    # start a solution search
    done_and_checked = False
    attempts = 0
    # DEBUGG show variables, bounds
    if self.debug_mode:
      self._debug_pyomo_print(m)
    while not done_and_checked:
      attempts += 1
      print(f'DEBUGG solve attempt {attempts} ...:')
      # solve
      # TODO someday if we want to give user access to options, we can add them to this dict. For now, no options.
      solve_options = {}
      soln = pyo.SolverFactory(self._solver).solve(m, options=solve_options)
      # check solve status
      if soln.solver.status == SolverStatus.ok and soln.solver.termination_condition == TerminationCondition.optimal:
        print('DEBUGG ... solve was successful!')
      else:
        print('DEBUGG ... solve was unsuccessful!')
        print('DEBUGG ... status:', soln.solver.status)
        print('DEBUGG ... termination:', soln.solver.termination_condition)
        self._debug_pyomo_print(m)
        print('Resource Map:')
        pprint.pprint(m.resource_index_map)
        raise RuntimeError
      # try validating
      print('DEBUGG ... validating ...')
      validation_errs = self.validate(m.Components, m.Activity, m.Times, meta)
      if validation_errs:
        done_and_checked = False
        print('DEBUGG ... validation concerns raised:')
        for e in validation_errs:
          print('DEBUGG ... ... Time {t} ({time}) Component "{c}" Resource "{r}": {m}'
                .format(t=e['time_index'],
                        time=e['time'],
                        c=e['component'].name,
                        r=e['resource'],
                        m=e['msg']))
          self._create_production_limit(m, e)
        # go back and solve again
        # raise NotImplementedError('Validation failed, but idk how to handle that yet')
      else:
        print('DEBUGG Solve successful and no validation concerns raised.')
        done_and_checked = True
      if attempts > 100:
        raise RuntimeError('Exceeded validation attempt limit!')
    if self.debug_mode:
      soln.write()
      self._debug_print_soln(m)
    # return dict of numpy arrays
    result = self._retrieve_solution(m)
    return result

  def check_converged(self, new, old, components):
    """
      Checks convergence of consecutive dispatch solves
      @ In, new, dict, results of dispatch # TODO should this be the model rather than dict?
      @ In, old, dict, results of previous dispatch
      @ In, components, list, HERON component list
      @ Out, converged, bool, True if convergence is met
    """
    tol = 1e-4 # TODO user option
    if old is None:
      return False
    converged = True
    for comp in components:
      intr = comp.get_interaction()
      name = comp.name
      tracker = comp.get_tracking_vars()[0]
      if intr.is_governed(): # by "is_governed" we mean "isn't optimized in pyomo"
        # check activity L2 norm as a differ
        # TODO this may be specific to storage right now
        res = intr.get_resource()
        scale = np.max(old[name][tracker][res])
        diff = np.linalg.norm(new[name][tracker][res] - old[name][tracker][res]) / (scale if scale != 0 else 1)
        if diff > tol:
          converged = False
    return converged

  def needs_convergence(self, components):
    """
      Determines whether the current setup needs convergence to solve.
      @ In, components, list, HERON component list
      @ Out, needs_convergence, bool, True if iteration is needed
    """
    for comp in components:
      intr = comp.get_interaction()
      # storages with a prescribed strategy MAY need iteration
      if intr.is_governed():
        return True
    # if we get here, no iteration is needed
    return False


  ### PYOMO Element Constructors
  def _create_production_limit(self, m, validation):
    """
      Creates pyomo production constraint given validation errors
      @ In, m, pyo.ConcreteModel, associated model
      @ In, validation, dict, information from Validator about limit violation
      @ Out, None
    """
    # TODO could validator write a symbolic expression on request? That'd be sweet.
    comp = validation['component']
    resource = validation['resource']
    r = m.resource_index_map[comp][resource]
    t = validation['time_index']
    limit = validation['limit']
    limits = np.zeros(len(m.Times))
    limits[t] = limit
    limit_type = validation['limit_type']
    prod_name = f'{comp.name}_production'
    rule = lambda mod: self._prod_limit_rule(prod_name, r, limits, limit_type, t, mod)
    constr = pyo.Constraint(rule=rule)
    counter = 1
    name_template = '{c}_{r}_{t}_vld_limit_constr_{{i}}'.format(c=comp.name,
                                                                r=resource,
                                                                t=t)
    # make sure we get a unique name for this constraint
    name = name_template.format(i=counter)
    while getattr(m, name, None) is not None:
      counter += 1
      name = name_template.format(i=counter)
    setattr(m, name, constr)
    print(f'DEBUGG added validation constraint "{name}"')

  def _create_production_param(self, m, comp, values, tag=None):
    """
      Creates production pyomo fixed parameter object for a component
      @ In, m, pyo.ConcreteModel, associated model
      @ In, comp, HERON Component, component to make production variables for
      @ In, values, np.array(float), values to set for param
      @ In, tag, str, optional, if not None then name will be component_[tag]
      @ Out, prod_name, str, name of production variable
    """
    name = comp.name
    if tag is None:
      tag = 'production'
    # create pyomo indexer for this component's resources
    res_indexer = pyo.Set(initialize=range(len(m.resource_index_map[comp])))
    setattr(m, f'{name}_res_index_map', res_indexer)
    prod_name = f'{name}_{tag}'
    init = (((0, t), values[t]) for t in m.T)
    prod = pyo.Param(res_indexer, m.T, initialize=dict(init))
    setattr(m, prod_name, prod)
    return prod_name

  def _create_production(self, m, comp, meta):
    """
      Creates all pyomo variable objects for a non-storage component
      @ In, m, pyo.ConcreteModel, associated model
      @ In, comp, HERON Component, component to make production variables for
      @ In, meta, dict, dictionary of state variables
      @ Out, None
    """
    prod_name = self._create_production_variable(m, comp, meta)
    ## if you cannot set limits directly in the production variable, set separate contraint:
    ## Method 1: set variable bounds directly --> TODO more work needed, but would be nice
    # lower, upper = self._get_prod_bounds(m, comp)
    # limits should be None unless specified, so use "getters" from dictionaries
    # bounds = lambda m, r, t: (lower.get(r, None), upper.get(r, None))
    ## Method 2: set variable bounds directly --> TODO more work needed, but would be nice
    # self._create_capacity(m, comp, prod_name, meta)    # capacity constraints
    # transfer function governs input -> output relationship
    self._create_transfer(m, comp, prod_name)
    # ramp rates TODO ## INCLUDING previous-time boundary condition TODO

  def _create_production_variable(self, m, comp, meta, tag=None, add_bounds=True, **kwargs):
    """
      Creates production pyomo variable object for a component
      @ In, m, pyo.ConcreteModel, associated model
      @ In, comp, HERON Component, component to make production variables for
      @ In, tag, str, optional, if not None then name will be component_[tag]; otherwise "production"
      @ In, add_bounds, bool, optional, if True then determine and set bounds for variable
      @ In, kwargs, dict, optional, passalong kwargs to pyomo variable
      @ Out, prod_name, str, name of production variable
    """
    if tag is None:
      tag = 'production'
    name = comp.name
    cap_res = comp.get_capacity_var()       # name of resource that defines capacity
    limit_r = m.resource_index_map[comp][cap_res] # production index of the governing resource
    # create pyomo indexer for this component's resources
    indexer_name = f'{name}_res_index_map'
    indexer = getattr(m, indexer_name, None)
    if indexer is None:
      indexer = pyo.Set(initialize=range(len(m.resource_index_map[comp])))
      setattr(m, indexer_name, indexer)
    prod_name = f'{name}_{tag}'
    caps, mins = self._find_production_limits(m, comp, meta)
    if min(caps) < 0:
      # we have a unit that's consuming, so we need to flip the variables to be sensible
      mins, caps = caps, mins
      inits = caps
    else:
      inits = mins
    if add_bounds:
      # create bounds based in min, max operation
      bounds = lambda m, r, t: (mins[t] if r == limit_r else None, caps[t] if r == limit_r else None)
      initial = lambda m, r, t: inits[t] if r == limit_r else 0
    else:
      bounds = (None, None)
      initial = 0
    # production variable depends on resources, time
    #FIXME initials! Should be lambda with mins for tracking var!
    prod = pyo.Var(indexer, m.T, initialize=initial, bounds=bounds, **kwargs)
    # TODO it may be that we need to set variable values to avoid problems in some solvers.
    # if comp.is_dispatchable() == 'fixed':
    #   for t, _ in enumerate(m.Times):
    #     prod[limit_r, t].fix(caps[t])
    setattr(m, prod_name, prod)
    return prod_name

  def _create_capacity_constraints(self, m, comp, prod_name, meta):
    """
      Creates pyomo capacity constraints
      @ In, m, pyo.ConcreteModel, associated model
      @ In, comp, HERON Component, component to make variables for
      @ In, prod_name, str, name of production variable
      @ In, meta, dict, additional state information
      @ Out, None
    """
    cap_res = comp.get_capacity_var()       # name of resource that defines capacity
    r = m.resource_index_map[comp][cap_res] # production index of the governing resource
    caps, mins = self._find_production_limits(m, comp, meta)
    # capacity
    max_rule = lambda mod, t: self._capacity_rule(prod_name, r, caps, mod, t)
    constr = pyo.Constraint(m.T, rule=max_rule)
    setattr(m, '{c}_{r}_capacity_constr'.format(c=comp.name, r=cap_res), constr)
    # minimum
    min_rule = lambda mod, t: self._min_prod_rule(prod_name, r, caps, mins, mod, t)
    constr = pyo.Constraint(m.T, rule=min_rule)
    # set initial conditions
    for t, time in enumerate(m.Times):
      cap = caps[t]
      if cap == mins[t]:
        # initialize values so there's no boundary errors
        var = getattr(m, prod_name)
        values = var.get_values()
        for k in values:
          values[k] = cap
        var.set_values(values)
    setattr(m, '{c}_{r}_minprod_constr'.format(c=comp.name, r=cap_res), constr)

  def _find_production_limits(self, m, comp, meta):
    """
      Determines the capacity limits of a unit's operation, in time.
      @ In, m, pyo.ConcreteModel, associated model
      @ In, comp, HERON Component, component to make variables for
      @ In, meta, dict, additional state information
      @ Out, caps, array, max production values by time
      @ Out, mins, array, min production values by time
    """
    cap_res = comp.get_capacity_var()       # name of resource that defines capacity
    r = m.resource_index_map[comp][cap_res] # production index of the governing resource
    # production is always lower than capacity
    ## NOTE get_capacity returns (data, meta) and data is dict
    ## TODO does this work with, e.g., ARMA-based capacities?
    ### -> "time" is stored on "m" and could be used to correctly evaluate the capacity
    caps = []
    mins = []
    for t, time in enumerate(m.Times):
      meta['HERON']['time_index'] = t + m.time_offset
      cap = comp.get_capacity(meta)[0][cap_res] # value of capacity limit (units of governing resource)
      caps.append(cap)
      if (comp.is_dispatchable() == 'fixed'):
        minimum = cap
      else:
        minimum = comp.get_minimum(meta)[0][cap_res]
      mins.append(minimum)
    return caps, mins

  def _create_transfer(self, m, comp, prod_name):
    """
      Creates pyomo transfer function constraints
      @ In, m, pyo.ConcreteModel, associated model
      @ In, comp, HERON Component, component to make variables for
      @ In, prod_name, str, name of production variable
      @ Out, None
    """
    name = comp.name
    # transfer functions
    # e.g. 2A + 3B -> 1C + 2E
    # get linear coefficients
    # TODO this could also take a transfer function from an external Python function assuming
    #    we're careful about how the expression-vs-float gets used
    #    and figure out how to handle multiple ins, multiple outs
    ratios = self._get_transfer_coeffs(m, comp)
    ref_r, ref_name, _ = ratios.pop('__reference', (None, None, None))
    for resource, ratio in ratios.items():
      r = m.resource_index_map[comp][resource]
      rule_name = '{c}_{r}_{fr}_transfer'.format(c=name, r=resource, fr=ref_name)
      rule = lambda mod, t: self._transfer_rule(ratio, r, ref_r, prod_name, mod, t)
      constr = pyo.Constraint(m.T, rule=rule)
      setattr(m, rule_name, constr)

  def _create_storage(self, m, comp, initial_storage, meta):
    """
      Creates storage pyomo variable objects for a storage component
      Similar to create_production, but for storages
      @ In, m, pyo.ConcreteModel, associated model
      @ In, comp, HERON Component, component to make production variables for
      @ In, initial_storage, dict, initial storage levels
      @ In, meta, dict, additional state information
      @ Out, level_name, str, name of storage level variable
    """
    prefix = comp.name
    # what resource index? Isn't it always 0? # assumption
    r = 0 # NOTE this is only true if each storage ONLY uses 1 resource
    # storages require a few variables:
    # (1) a level tracker,
    level_name = self._create_production_variable(m, comp, meta, tag='level')
    # -> set operational limits
    # self._create_capacity(m, comp, level_name, meta)
    # (2, 3) separate charge/discharge trackers, so we can implement round-trip efficiency and ramp rates
    charge_name = self._create_production_variable(m, comp, meta, tag='charge', add_bounds=False, within=pyo.NonPositiveReals)
    discharge_name = self._create_production_variable(m, comp, meta, tag='discharge', add_bounds=False, within=pyo.NonNegativeReals)
    # balance level, charge/discharge
    level_rule_name = prefix + '_level_constr'
    rule = lambda mod, t: self._level_rule(comp, level_name, charge_name, discharge_name,
                                           initial_storage, r, mod, t)
    setattr(m, level_rule_name, pyo.Constraint(m.T, rule=rule))
    # (4) a binary variable to track whether we're charging or discharging, to prevent BOTH happening
    # -> 0 is charging, 1 is discharging
    # -> TODO make this a user-based option to disable, if they want to allow dual operation
    # -> -> but they should really think about if that's what they want!
    # FIXME currently introducing the bigM strategy also makes solves numerically unstable,
    # and frequently results in spurious errors. For now, disable it.
    allow_both = True # allow simultaneous charging and discharging
    if not allow_both:
      bin_name = self._create_production_variable(m, comp, meta, tag='dcforcer', add_bounds=False, within=pyo.Binary)
      # we need a large epsilon, but not so large that addition stops making sense
      # -> we don't know what any values for this component will be! How do we choose?
      # -> NOTE that choosing this value has VAST impact on solve stability!!
      large_eps = 1e8 #0.01 * sys.float_info.max
      # charging constraint: don't charge while discharging (note the sign matters)
      charge_rule_name = prefix + '_charge_constr'
      rule = lambda mod, t: self._charge_rule(charge_name, bin_name, large_eps, r, mod, t)
      setattr(m, charge_rule_name, pyo.Constraint(m.T, rule=rule))
      discharge_rule_name = prefix + '_discharge_constr'
      rule = lambda mod, t: self._discharge_rule(discharge_name, bin_name, large_eps, r, mod, t)
      setattr(m, discharge_rule_name, pyo.Constraint(m.T, rule=rule))

  def _create_conservation(self, m, resources, initial_storage, meta):
    """
      Creates pyomo conservation constraints
      @ In, m, pyo.ConcreteModel, associated model
      @ In, resources, list, list of resources in problem
      @ In, initial_storage, dict, initial storage levels
      @ In, meta, dict, dictionary of state variables
      @ Out, None
    """
    for res, resource in enumerate(resources):
      rule = lambda mod, t: self._conservation_rule(initial_storage, meta, resource, mod, t)
      constr = pyo.Constraint(m.T, rule=rule)
      setattr(m, '{r}_conservation'.format(r=resource), constr)

  def _create_objective(self, meta, m):
    """
      Creates pyomo objective function
      @ In, meta, dict, additional variables to pass through
      @ In, m, pyo.ConcreteModel, associated model
      @ Out, None
    """
    # cashflow eval
    rule = lambda mod: self._cashflow_rule(meta, mod)
    m.obj = pyo.Objective(rule=rule, sense=pyo.maximize)

  ### UTILITIES for general use
  def _get_prod_bounds(self, m, comp, meta):
    """
      Determines the production limits of the given component
      @ In, comp, HERON component, component to get bounds of
      @ In, meta, dict, additional state information
      @ Out, lower, dict, lower production limit (might be negative for consumption)
      @ Out, upper, dict, upper production limit (might be 0 for consumption)
    """
    raise NotImplementedError
    cap_res = comp.get_capacity_var()       # name of resource that defines capacity
    r = m.resource_index_map[comp][cap_res]
    maxs = []
    mins = []
    for t, time in enumerate(m.Times):
      meta['HERON']['time_index'] = t + m.time_offset
      cap = comp.get_capacity(meta)[0][cap_res]
      low = comp.get_minimum(meta)[0][cap_res]
      maxs.append(cap)
      if (comp.is_dispatchable() == 'fixed') or (low == cap):
        low = cap
        # initialize values to avoid boundary errors
        var = getattr(m, prod_name)
        values = var.get_values()
        for k in values:
          values[k] = cap
        var.set_values(values)
      mins.append(low)
    maximum = comp.get_capacity(None, None, None, None)[0][cap_res]
    # TODO minimum!
    # producing or consuming the defining resource?
    # -> put these in dictionaries so we can "get" limits or return None
    if maximum > 0:
      lower = {r: 0}
      upper = {r: maximum}
    else:
      lower = {r: maximum}
      upper = {r: 0}
    return lower, upper

  def _get_transfer_coeffs(self, m, comp):
    """
      Obtains transfer function ratios (assuming Linear ValuedParams)
      Form: 1A + 3B -> 2C + 4D
      Ratios are calculated with respect to first resource listed, so e.g. B = 3/1 * A
      TODO I think we can handle general external functions, maybe?
      @ In, m, pyo.ConcreteModel, associated model
      @ In, comp, HERON component, component to get coefficients of
      @ Out, ratios, dict, ratios of transfer function variables
    """
    name = comp.name
    transfer = comp.get_interaction().get_transfer()  # get the transfer ValuedParam, if any
    if transfer is None:
      return {}
    coeffs = transfer.get_coefficients() # linear transfer coefficients, dict as {resource: coeff}, SIGNS MATTER
    # it's all about ratios -> store as ratio of resource / first resource (arbitrary)
    first_r = None
    first_name = None
    first_coef = None
    ratios = {}
    for resource, coef in coeffs.items():
      r = m.resource_index_map[comp][resource]
      if first_r is None:
        # set up nominal resource to compare to
        first_r = r
        first_name = resource
        first_coef = coef
        ratios['__reference'] = (first_r, first_name, first_coef)
        continue
      ratio = coef / first_coef
      ratios[resource] = ratio
    return ratios

  def _retrieve_solution(self, m):
    """
      Extracts solution from Pyomo optimization
      @ In, m, pyo.ConcreteModel, associated (solved) model
      @ Out, result, dict, {comp: {activity: {resource: [production]}} e.g. generator[production][steam]
    """
    result = {} # {component: {activity_tag: {resource: production}}}
    for comp in m.Components:
      result[comp.name] = {}
      for tag in comp.get_tracking_vars():
        tag_values = self._retrieve_value_from_model(m, comp, tag)
        result[comp.name][tag] = tag_values
    return result

  def _retrieve_value_from_model(self, m, comp, tag):
    """
      Retrieve values of a series from the pyomo model.
      @ In, m, pyo.ConcreteModel, associated (solved) model
      @ In, comp, Component, relevant component
      @ In, tag, str, particular history type to retrieve
      @ Out, result, dict, {resource: [array], etc}
    """
    result = {}
    prod = getattr(m, f'{comp.name}_{tag}')
    if isinstance(prod, pyo.Var):
      kind = 'Var'
    elif isinstance(prod, pyo.Param):
      kind = 'Param'
    for res, comp_r in m.resource_index_map[comp].items():
      if kind == 'Var':
        result[res] = np.fromiter((prod[comp_r, t].value for t in m.T), dtype=float, count=len(m.T))
      elif kind == 'Param':
        result[res] = np.fromiter((prod[comp_r, t] for t in m.T), dtype=float, count=len(m.T))
    return result

  ### RULES for lambda function calls
  # these get called using lambda functions to make Pyomo constraints, vars, objectives, etc

  def _charge_rule(self, charge_name, bin_name, large_eps, r, m, t):
    """
      Constructs pyomo don't-charge-while-discharging constraints.
      For storage units specificially.
      @ In, charge_name, str, name of charging variable
      @ In, bin_name, str, name of forcing binary variable
      @ In, large_eps, float, a large-ish number w.r.t. storage activity
      @ In, r, int, index of stored resource (is this always 0?)
      @ In, m, pyo.ConcreteModel, associated model
      @ In, t, int, time index for capacity rule
      @ Out, rule, bool, inequality used to limit charge behavior
    """
    charge_var = getattr(m, charge_name)
    bin_var = getattr(m, bin_name)
    return -charge_var[r, t] <= (1 - bin_var[r, t]) * large_eps

  def _discharge_rule(self, discharge_name, bin_name, large_eps, r, m, t):
    """
      Constructs pyomo don't-discharge-while-charging constraints.
      For storage units specificially.
      @ In, discharge_name, str, name of discharging variable
      @ In, bin_name, str, name of forcing binary variable
      @ In, large_eps, float, a large-ish number w.r.t. storage activity
      @ In, r, int, index of stored resource (is this always 0?)
      @ In, m, pyo.ConcreteModel, associated model
      @ In, t, int, time index for capacity rule
      @ Out, rule, bool, inequality used to limit discharge behavior
    """
    discharge_var = getattr(m, discharge_name)
    bin_var = getattr(m, bin_name)
    return discharge_var[r, t] <= bin_var[r, t] * large_eps

  def _level_rule(self, comp, level_name, charge_name, discharge_name, initial_storage, r, m, t):
    """
      Constructs pyomo charge-discharge-level balance constraints.
      For storage units specificially.
      @ In, comp, Component, storage component of interest
      @ In, level_name, str, name of level-tracking variable
      @ In, charge_name, str, name of charging variable
      @ In, discharge_name, str, name of discharging variable
      @ In, r, int, index of stored resource (is this always 0?)
      @ In, m, pyo.ConcreteModel, associated model
      @ In, t, int, time index for capacity rule
      @ Out, rule, bool, inequality used to limit level behavior
    """
    level_var = getattr(m, level_name)
    charge_var = getattr(m, charge_name)
    discharge_var = getattr(m, discharge_name)
    if t > 0:
      previous = level_var[r, t-1]
      dt = m.Times[t] - m.Times[t-1]
    else:
      previous = initial_storage[comp]
      dt = m.Times[1] - m.Times[0]
    rte2 = comp.get_sqrt_RTE() # square root of the round-trip efficiency
    production = - rte2 * charge_var[r, t] - discharge_var[r, t] / rte2
    return level_var[r, t] == previous + production * dt

  def _capacity_rule(self, prod_name, r, caps, m, t):
    """
      Constructs pyomo capacity constraints.
      @ In, prod_name, str, name of production variable
      @ In, r, int, index of resource for capacity constraining
      @ In, caps, list(float), value to constrain resource at in time
      @ In, m, pyo.ConcreteModel, associated model
      @ In, t, int, time index for capacity rule
    """
    kind = 'lower' if min(caps) < 0 else 'upper'
    return self._prod_limit_rule(prod_name, r, caps, kind, t, m)

  def _prod_limit_rule(self, prod_name, r, limits, kind, t, m):
    """
      Constructs pyomo production constraints.
      @ In, prod_name, str, name of production variable
      @ In, r, int, index of resource for capacity constraining
      @ In, limits, list(float), values in time at which to constrain resource production
      @ In, kind, str, either 'upper' or 'lower' for limiting production
      @ In, t, int, time index for production rule (NOTE not pyomo index, rather fixed index)
      @ In, m, pyo.ConcreteModel, associated model
    """
    prod = getattr(m, prod_name)
    if kind == 'lower':
      # production must exceed value
      return prod[r, t] >= limits[t]
    elif kind == 'upper':
      return prod[r, t] <= limits[t]
    else:
      raise TypeError('Unrecognized production limit "kind":', kind)

  def _cashflow_rule(self, meta, m):
    """
      Objective function rule.
      @ In, meta, dict, additional variable passthrough
      @ In, m, pyo.ConcreteModel, associated model
      @ Out, total, float, evaluation of cost
    """
    activity = m.Activity # dict((comp, getattr(m, f"{comp.name}_production")) for comp in m.Components)
    state_args = {'valued': False}
    total = self._compute_cashflows(m.Components, activity, m.Times, meta,
                                    state_args=state_args, time_offset=m.time_offset)
    return total

  def _conservation_rule(self, initial_storage, meta, res, m, t):
    """
      Constructs conservation constraints.
      @ In, initial_storage, dict, initial storage levels at t==0 (not t+offset==0)
      @ In, meta, dict, dictionary of state variables
      @ In, res, str, name of resource
      @ In, m, pyo.ConcreteModel, associated model
      @ In, t, int, index of time variable
      @ Out, conservation, bool, balance check
    """
    balance = 0 # sum of production rates, which needs to be zero
    for comp, res_dict in m.resource_index_map.items():
      if res in res_dict:
        # activity information depends on if storage or component
        r = res_dict[res]
        intr = comp.get_interaction()
        if intr.is_type('Storage'):
          # Storages have 3 variables: level, charge, and discharge
          # -> so calculate activity
          charge = getattr(m, f'{comp.name}_charge')
          discharge = getattr(m, f'{comp.name}_discharge')
          # note that "charge" is negative (as it's consuming) and discharge is positive
          # -> so the intuitive |discharge| - |charge| becomes discharge + charge
          production = discharge[r, t] + charge[r, t]
        else:
          var = getattr(m, f'{comp.name}_production')
          # TODO move to this? balance += m._activity.get_activity(comp, res, t)
          production = var[r, t]
        balance += production
    return balance == 0 # TODO tol?

  def _min_prod_rule(self, prod_name, r, caps, minimums, m, t):
    """
      Constructs minimum production constraint
      @ In, prod_name, str, name of production variable
      @ In, r, int, index of resource for capacity constraining
      @ In, caps, list(float), capacity(t) value for component
      @ In, minimums, list(float), minimum allowable production in time
      @ In, m, pyo.ConcreteModel, associated model
      @ In, t, int, index of time variable
      @ Out, minimum, bool, min check
    """
    prod = getattr(m, prod_name)
    # negative capacity means consuming instead of producing
    if max(caps) > 0:
      return prod[r, t] >= minimums[t]
    else:
      return prod[r, t] <= minimums[t]

  def _transfer_rule(self, ratio, r, ref_r, prod_name, m, t):
    """
      Constructs transfer function constraints
      @ In, ratio, float, ratio for resource to nominal first resource
      @ In, r, int, index of transfer resource
      @ In, ref_r, int, index of reference resource
      @ In, prod_name, str, name of production variable
      @ In, m, pyo.ConcreteModel, associated model
      @ In, t, int, index of time variable
      @ Out, transfer, bool, transfer ratio check
    """
    prod = getattr(m, prod_name)
    return prod[r, t] == prod[ref_r, t] * ratio # TODO tolerance??

  ### DEBUG
  def _debug_pyomo_print(self, m):
    """
      Prints the setup pieces of the Pyomo model
      @ In, m, pyo.ConcreteModel, model to interrogate
      @ Out, None
    """
    print('/' + '='*80)
    print('DEBUGG model pieces:')
    print('  -> objective:')
    print('     ', m.obj.pprint())
    print('  -> variables:')
    for var in m.component_objects(pyo.Var):
      print('     ', var.pprint())
    print('  -> constraints:')
    for constr in m.component_objects(pyo.Constraint):
      print('     ', constr.pprint())
    print('\\' + '='*80)
    print('')

  def _debug_print_soln(self, m):
    """
      Prints the solution from the Pyomo model
      @ In, m, pyo.ConcreteModel, model to interrogate
      @ Out, None
    """
    print('*'*80)
    print('DEBUGG solution:')
    print('  objective value:', m.obj())
    for c, comp in enumerate(m.Components):
      name = comp.name
      print('  component:', c, name)
      for tracker in comp.get_tracking_vars():
        for res, r in m.resource_index_map[comp].items():
          print(f'    tracker: {tracker} resource {r}: {res}')
          for t, time in enumerate(m.Times):
            prod = getattr(m, f'{name}_{tracker}')
            if isinstance(prod, pyo.Var):
              print('      time:', t + m.time_offset, time, prod[r, t].value)
            elif isinstance(prod, pyo.Param):
              print('      time:', t + m.time_offset, time, prod[r, t])
    print('*'*80)





# DispatchState for Pyomo dispatcher
class PyomoState(DispatchState):
  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    DispatchState.__init__(self)
    self._model = None # Pyomo model object

  def initialize(self, components, resources_map, times, model):
    """
      Connect information about this State to other objects
      @ In, components, list, HERON components
      @ In, resources_map, dict, map of component names to resources used
      @ In, times, np.array, values of "time" this state represents
      @ In, model, pyomo.Model, associated model for this state
      @ Out, None
    """
    DispatchState.initialize(self, components, resources_map, times)
    self._model = model

  def get_activity_indexed(self, comp, activity, r, t, valued=True, **kwargs):
    """
      Getter for activity level.
      @ In, comp, HERON Component, component whose information should be retrieved
      @ In, activity, str, tracking variable name for activity subset
      @ In, r, int, index of resource to retrieve (as given by meta[HERON][resource_indexer])
      @ In, t, int, index of time at which activity should be provided
      @ In, valued, bool, optional, if True then get float value instead of pyomo expression
      @ In, kwargs, dict, additional pass-through keyword arguments
      @ Out, activity, float, amount of resource "res" produced/consumed by "comp" at time "time";
                              note positive is producting, negative is consuming
    """
    prod = getattr(self._model, f'{comp.name}_{activity}')[r, t]
    if valued:
      return prod()
    return prod

  def set_activity_indexed(self, comp, r, t, value, valued=False):
    raise NotImplementedError




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED

from ravenframework.utils import InputData, InputTypes
from ravenframework.EntityFactoryBase import EntityFactory

from .SyntheticHistory import SyntheticHistory
from .StaticHistory import StaticHistory
from .ROM import ROM
from .Function import Function
from .Parametric import Parametric, FixedValue, OptBounds, SweepValues
from .Linear import Linear
from .Variable import Variable
from .Activity import Activity

class ValuedParamFactory(EntityFactory):
  """
    Factory for ValuedParams
  """
  def __init__(self, *args):
    """
      Instantiate
      @ In, args, list, positional arguments
      @ Out, None
    """
    super().__init__(*args)
    # TODO registered_entities = [] # for filling in values

  def make_input_specs(self, name, descr=None, allowed=None, kind='singular'):
    """
      Fill input specs for the provided name and description.
      @ In, name, str, name of new spec
      @ In, descr, str, optional, description of spec
      @ In, allowed, list, optional, string list of allowable types of ValuedParam. Overrides "kind".
      @ In, kind, str, optional, kind of ValuedParam grouping (default)
      @ Out, spec, InputData, specification
    """
    add_descr = ''
    if allowed is None:
      allowed = allowable[kind]
      allowed_str = ', '.join(['\\xmlNode{{{}}}'.format(a) for a in allowed])
      add_descr = rf"""This value can be taken from any \emph{{one}} of the sources as subnodes (described below): {allowed_str}."""

    if descr is None:
      description = add_descr
    else:
      description = descr + r"""\\ \\""" + add_descr

    spec = InputData.parameterInputFactory(name, descr=description)
    for typ, klass in self._registeredTypes.items():
      if typ in allowed:
        spec.addSub(klass.get_input_specs())
        # addons
    spec.addSub(
      InputData.parameterInputFactory(
        'multiplier',
        contentType=InputTypes.FloatType,
        descr=r"""Multiplies any value obtained by this parameter by the given value. \default{1}"""
      )
    )
    return spec

factory = ValuedParamFactory('ValuedParam')

# fixed in inner
factory.registerType('fixed_value', FixedValue)
factory.registerType('sweep_values', SweepValues)
factory.registerType('opt_bounds', OptBounds)
factory.registerType('variable', Variable)
# frequent revaluation
factory.registerType('CSV', StaticHistory)
factory.registerType('ARMA', SyntheticHistory)
factory.registerType('ROM', ROM)
factory.registerType('Function', Function)
factory.registerType('activity', Activity)
# ratios, transfers
factory.registerType('linear', Linear)
# TODO add: ROM

# TODO are transfer functions and valued evaluations really the same creature?

# map of "kinds" of ValuedParams to the default acceptable ValuedParam types
allowable = {}
# transfer functions, such as producing components' transfer functions
allowable['transfer'] = ['linear', 'Function']
# single evaluations, like cashflow prices and component capacities
allowable['singular'] = [
  'fixed_value',
  'sweep_values',
  'opt_bounds',
  'variable',
  'ARMA',
  'Function',
  'ROM',
  'CSV'
]
# evaluations available only after dispatch (e.g. for economics)
## for example, we can't base a capacity on the dispatch activity ... right?
allowable['post-dispatch'] = allowable['singular'] + ['activity']
allowable['all'] = list(factory.knownTypes())




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
import os
import sys
import dill as pk

def load_heron_lib(path, retry=0):
  """
    Loads serialized heron file
    @ In, path, str, path to file to load from
    @ In, retry, int, number of re-attempts for finding file
    @ Out, case, Case, HERON case object (None if not found)
    @ Out, components, list, list of HERON component objects (None if not found)
    @ Out, sources, list, list of HERON source objects (None if not found)
  """
  found = False
  while not found:
    counter = 0
    try:
      with open(path, 'rb') as lib:
        external_funcs = os.path.abspath(os.path.join(os.path.curdir, '..'))
        sys.path.append(external_funcs)
        case, components, sources = pk.load(lib)
        found = True
    except FileNotFoundError:
      print('WARNING: "{n}" not yet found; waiting and retrying {r} times ...'
            .format(n=path, r=retry-counter))
      counter += 1
      if counter > retry:
        case, components, source = None, None, None
        break
  return case, components, sources




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the Component entity.
"""
from __future__ import unicode_literals, print_function
import sys
from collections import defaultdict
import numpy as np
from HERON.src.base import Base
import xml.etree.ElementTree as ET
from HERON.src.Economics import CashFlowUser
from HERON.src.ValuedParams import factory as vp_factory
from HERON.src.ValuedParamHandler import ValuedParamHandler
from HERON.src import _utils as hutils

try:
  import ravenframework
except ModuleNotFoundError:
  framework_path = hutils.get_raven_loc()
  sys.path.append(framework_path)
from ravenframework.utils import InputData, xmlUtils,InputTypes

# TODO can we use EntityFactory from RAVEN?
def factory(xml, method='sweep'):
  """
    Tool for constructing compnents without the input_loader
    TODO can this be set up so the input_loader calls it instead of the methods directly?
    @ In, xml, ET.Element, node from which to read component settings
    @ In, method, string, optional, operational mode for case
    @ Out, comp, Component instance, component constructed
  """
  comp = Component()
  comp.read_input(xml, method)
  return comp

class Component(Base, CashFlowUser):
  """
    Represents a unit in the grid analysis. Each component has a single "interaction" that
    describes what it can do (produce, store, demand)
  """
  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, input_specs, InputData, specs
    """
    input_specs = InputData.parameterInputFactory('Component', ordered=False, baseNode=None,
        descr=r"""defines a component as an element of the grid system. Components are defined by the action they
              perform such as \xmlNode{produces} or \xmlNode{consumes}; see details below.""")
    input_specs.addParam('name', param_type=InputTypes.StringType, required=True,
        descr=r"""identifier for the component. This identifier will be used to generate variables
              and relate signals to this component throughout the HERON analysis.""")
    # production
    ## this unit may be able to make stuff, possibly from other stuff
    input_specs.addSub(Producer.get_input_specs())
    # storage
    ## this unit may be able to store stuff
    input_specs.addSub(Storage.get_input_specs())
    # demands
    ## this unit may have a certain demand that must be met
    input_specs.addSub(Demand.get_input_specs())
    # this unit probably has some economics
    input_specs = CashFlowUser.get_input_specs(input_specs)
    return input_specs

  def __init__(self, **kwargs):
    """
      Constructor
      @ In, kwargs, dict, optional, arguments to pass to other constructors
      @ Out, None
    """
    Base.__init__(self, **kwargs)
    CashFlowUser.__init__(self)
    self.name = None
    self._produces = []
    self._stores = []
    self._demands = []

  def __repr__(self):
    """
      String representation.
      @ In, None
      @ Out, __repr__, string representation
    """
    return '<HERON Component "{}">'.format(self.name)

  def read_input(self, xml, mode):
    """
      Sets settings from input file
      @ In, xml, xml.etree.ElementTree.Element, input from user
      @ In, mode, string, case mode to operate in (e.g. 'sweep' or 'opt')
      @ Out, None
    """
    # get specs for allowable inputs
    specs = self.get_input_specs()()
    specs.parseNode(xml)
    self.name = specs.parameterValues['name']
    self.raiseADebug('Loading component "{}"'.format(self.name))
    for item in specs.subparts:
      if self.get_interaction() and item.getName() in ['produces', 'stores', 'demands']:
        self.raiseAnError(NotImplementedError, 'Currently each Component can only have one interaction (produces, stores, demands)! Check Component "{}"'.format(self.name))
      # read in producers
      if item.getName() == 'produces':
        prod = Producer(messageHandler=self.messageHandler)
        try:
          prod.read_input(item, mode, self.name)
        except IOError as e:
          self.raiseAWarning('Errors while reading component "{}"!'.format(self.name))
          raise e
        self._produces.append(prod)
      # read in storages
      elif item.getName() == 'stores':
        store = Storage(messageHandler=self.messageHandler)
        store.read_input(item, mode, self.name)
        self._stores.append(store)
      # read in demands
      elif item.getName() == 'demands':
        demand = Demand(messageHandler=self.messageHandler)
        demand.read_input(item, mode, self.name)
        self._demands.append(demand)
      # read in economics
      elif item.getName() == 'economics':
        econ_node = item # need to read AFTER the interactions!
    # after looping over nodes, finish up
    if econ_node is None:
      self.raiseAnError(IOError, '<economics> node missing from component "{}"!'.format(self.name))
    CashFlowUser.read_input(self, econ_node)

  def get_crossrefs(self):
    """
      Collect the required value entities needed for this component to function.
      @ In, None
      @ Out, crossrefs, dict, mapping of dictionaries with information about the entities required.
    """
    inter = self.get_interaction()
    crossrefs = {inter: inter.get_crossrefs()}
    crossrefs.update(self._economics.get_crossrefs())
    return crossrefs

  def set_crossrefs(self, refs):
    """
      Connect cross-reference material from other entities to the ValuedParams in this component.
      @ In, refs, dict, dictionary of entity information
      @ Out, None
    """
    try_match = self.get_interaction()
    for interaction in list(refs.keys()):
      # find associated interaction
      if try_match == interaction:
        try_match.set_crossrefs(refs.pop(interaction))
        break
    # send what's left to the economics
    self._economics.set_crossrefs(refs)
    # if anything left, there's an issue
    assert not refs

  def get_interaction(self):
    """
      Return the interactions this component uses.
      TODO could this just return the only non-empty one, since there can only be one?
      @ In, None
      @ Out, interactions, list, list of Interaction instances
    """
    try:
      return (self._produces + self._stores + self._demands)[0]
    except IndexError: # there are no interactions!
      return None

  def get_sqrt_RTE(self):
    """
      Provide the square root of the round-trip efficiency for this component.
      Note we use the square root due to splitting loss across the input and output.
      @ In, None
      @ Out, RTE, float, round-trip efficiency as a multiplier
    """
    return self.get_interaction().get_sqrt_RTE()

  def print_me(self, tabs=0, tab='  '):
    """
      Prints info about self
      @ In, tabs, int, optional, number of tabs to insert before prints
      @ In, tab, str, optional, characters to use to denote hierarchy
      @ Out, None
    """
    pre = tab*tabs
    self.raiseADebug(pre+'Component:')
    self.raiseADebug(pre+'  name:', self.name)
    self.get_interaction().print_me(tabs=tabs+1, tab=tab)

  def get_inputs(self):
    """
      returns list of all resources consumed here
      @ In, None
      @ Out, inputs, set, set of input resources as strings (resources that are taken/consumed/stored)
    """
    inputs = set()
    # simply combine the inputs for the interaction
    inputs.update(self.get_interaction().get_inputs())
    return inputs

  def get_outputs(self):
    """
      returns list of all resources producable here
      @ In, None
      @ Out, outputs, set, set of output resources as strings (resources that are produced/provided)
    """
    outputs = set()
    outputs.update(self.get_interaction().get_outputs())
    return outputs

  def get_resources(self):
    """
      Provides the full set of resources used by this component.
      @ In, None
      @ Out, res, set, set(str) of resource names
    """
    res = set()
    res.update(self.get_inputs())
    res.update(self.get_outputs())
    return res

  def get_capacity(self, meta, raw=False):
    """
      returns the capacity of the interaction of this component
      @ In, meta, dict, arbitrary metadata from EGRET
      @ In, raw, bool, optional, if True then return the ValuedParam instance for capacity, instead of the evaluation
      @ Out, capacity, float (or ValuedParam), the capacity of this component's interaction
    """
    return self.get_interaction().get_capacity(meta, raw=raw)

  def get_minimum(self, meta, raw=False):
    """
      returns the minimum of the interaction of this component
      @ In, meta, dict, arbitrary metadata from EGRET
      @ In, raw, bool, optional, if True then return the ValuedParam instance for capacity, instead of the evaluation
      @ Out, capacity, float (or ValuedParam), the capacity of this component's interaction
    """
    return self.get_interaction().get_minimum(meta, raw=raw)

  def get_capacity_var(self):
    """
      Returns the variable that is used to define this component's capacity.
      @ In, None
      @ Out, var, str, name of capacity resource
    """
    return self.get_interaction().get_capacity_var()

  def get_tracking_vars(self):
    """
      Provides the variables used by this component to track dispatch
      @ In, None
      @ Out, get_tracking_vars, list, variable name list
    """
    return self.get_interaction().get_tracking_vars()

  def is_dispatchable(self):
    """
      Returns the dispatchability indicator of this component.
      TODO Note that despite the name, this is NOT boolean, but a string indicator.
      @ In, None
      @ Out, dispatchable, str, dispatchability (e.g. independent, dependent, fixed)
    """
    return self.get_interaction().is_dispatchable()

  def is_governed(self):
    """
      Determines if this component is optimizable or governed by some function.
      @ In, None
      @ Out, is_governed, bool, whether this component is governed.
    """
    return self.get_interaction().is_governed()

  def set_capacity(self, cap):
    """
      Set the float value of the capacity of this component's interaction
      @ In, cap, float, value
      @ Out, None
    """
    return self.get_interaction().set_capacity(cap)

  def produce(self, request, meta, raven_variables, dispatch, t, level=None):
    """
      Enacts the transfer function for this component to act based on a request.
      FIXME was used for "generic" dispatcher, does it still apply?
      @ In, request, dict, mapping of requested resource usage to amount requested (negative is
                           consume, positive is produce)
      @ In, meta, dict, metadata information for current status in run
      @ In, raven_variables, dict, variables from RAVEN TODO part of meta!
      @ In, dispatch, DispatchState, expression of the current activity levels in the system
      @ In, t, int, index of "time" at which this production should be performed
      @ In, level, float, for storages indicates the amount currently stored
      @ Out, balance, dict, full dict of resources used and produced for request
      @ Out, meta, dict, updated metadata dictionary
    """
    #balance = defaultdict(float)
    interaction = self.get_interaction()
    balance, meta = interaction.produce(request, meta, raven_variables, dispatch, t, level)
    #for resource, quantity in int_balance.items():
    #  balance[resource] += quantity
    return balance, meta

  def produce_max(self, meta, raven_variables, dispatch, t):
    """
      Determines the maximum production possible for this component.
      @ In, meta, dict, metadata information for current status in run
      @ In, raven_variables, dict, variables from RAVEN TODO part of meta!
      @ In, dispatch, DispatchState, expression of the current activity levels in the system
      @ In, t, int, index of "time" at which this production should be performed
      @ Out, balance, dict, full dict of resources used and produced for request
      @ Out, meta, dict, updated metadata dictionary
    """
    #balance = defaultdict(float)
    interaction = self.get_interaction()
    balance, meta = interaction.produce_max(meta, raven_variables, dispatch, t)
    #for resource, quantity in int_balance.items():
    #  balance[resource] += quantity
    return balance, meta

  def produce_min(self, meta, raven_variables, dispatch, t):
    """
      Determines the minimum production possible for this component.
      @ In, meta, dict, metadata information for current status in run
      @ In, raven_variables, dict, variables from RAVEN TODO part of meta!
      @ In, dispatch, DispatchState, expression of the current activity levels in the system
      @ In, t, int, index of "time" at which this production should be performed
      @ Out, balance, dict, full dict of resources used and produced for request
      @ Out, meta, dict, updated metadata dictionary
    """
    #balance = defaultdict(float)
    interaction = self.get_interaction()
    balance, meta = interaction.produce_min(meta, raven_variables, dispatch, t)
    #for resource, quantity in int_balance.items():
    #  balance[resource] += quantity
    return balance, meta

  def get_capacity_param(self):
    """
      Provides direct access to the ValuedParam for the capacity of this component.
      @ In, None
      @ Out, cap, ValuedParam, capacity valued param
    """
    intr = self.get_interaction()
    return intr.get_capacity(None, None, None, None, raw=True)







class Interaction(Base):
  """
    Base class for component interactions (e.g. Producer, Storage, Demand)
  """
  tag = 'interacts' # node name in input file

  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, input_specs, InputData, specs
    """
    if cls.tag == 'produces':
      desc = r"""indicates that this component produces one or more resources by consuming other resources."""
      resource_desc = r"""the resource produced by this component's activity."""
    elif cls.tag == 'stores':
      desc = r"""indicates that this component stores one resource, potentially absorbing or providing that resource."""
      resource_desc = r"""the resource stored by this component."""
    elif cls.tag == "demands":
      desc = r"""indicates that this component exclusively consumes a resource."""
      resource_desc = r"""the resource consumed by this component."""
    specs = InputData.parameterInputFactory(cls.tag, ordered=False, descr=desc)
    specs.addParam('resource', param_type=InputTypes.StringListType, required=True,
        descr=resource_desc)
    dispatch_opts = InputTypes.makeEnumType('dispatch_opts', 'dispatch_opts', ['fixed', 'independent', 'dependent'])
    specs.addParam('dispatch', param_type=dispatch_opts,
        descr=r"""describes the way this component should be dispatched, or its flexibility.
              \texttt{fixed} indicates the component always fully dispatched at its maximum level.
              \texttt{independent} indicates the component is fully dispatchable by the dispatch optimization algorithm.
              \texttt{dependent} indicates that while this component is not directly controllable by the dispatch
              algorithm, it can however be flexibly dispatched in response to other units changing dispatch level.
              For example, when attempting to increase profitability, the \texttt{fixed} components are not adjustable,
              but the \texttt{independent} components can be adjusted to attempt to improve the economic metric.
              In response to the \texttt{independent} component adjustment, the \texttt{dependent} components
              may respond to balance the resource usage from the changing behavior of other components.""")

    descr = r"""the maximum value at which this component can act, in units corresponding to the indicated resource. """
    cap = vp_factory.make_input_specs('capacity', descr=descr)
    cap.addParam('resource', param_type=InputTypes.StringType,
        descr=r"""indicates the resource that defines the capacity of this component's operation. For example,
              if a component consumes steam and electricity to produce hydrogen, the capacity of the component
              can be defined by the maximum steam consumable, maximum electricity consumable, or maximum
              hydrogen producable. Any choice should be nominally equivalent, but determines the units
              of the value of this node.""")
    specs.addSub(cap)

    descr = r"""provides the minimum value at which this component can act, in units of the indicated resource. """
    minn = vp_factory.make_input_specs('minimum', descr=descr)
    minn.addParam('resource', param_type=InputTypes.StringType,
        descr=r"""indicates the resource that defines the minimum activity level for this component,
              as with the component's capacity.""")
    specs.addSub(minn)

    return specs

  def __init__(self, **kwargs):
    """
      Constructor
      @ In, kwargs, dict, arbitrary pass-through arguments
      @ Out, None
    """
    Base.__init__(self, **kwargs)
    self._capacity = None               # upper limit of this interaction
    self._capacity_var = None           # which variable limits the capacity (could be produced or consumed?)
    self._signals = set()               # dependent signals for this interaction
    self._crossrefs = defaultdict(dict) # crossrefs objects needed (e.g. armas, etc), as {attr: {tag, name, obj})
    self._dispatchable = None           # independent, dependent, or fixed?
    self._minimum = None                # lowest interaction level, if dispatchable
    self._minimum_var = None            # limiting variable for minimum
    self._function_method_map = {}      # maps things that call functions to the method within the function that needs calling
    self._transfer = None               # the production rate (if any), in produces per consumes
                                        #   for example, {(Producer, 'capacity'): 'method'}
    self._sqrt_rte = 1.0                # sqrt of the round-trip efficiency for this interaction
    self._tracking_vars = []            # list of trackable variables for dispatch activity

  def read_input(self, specs, mode, comp_name):
    """
      Sets settings from input file
      @ In, specs, InputData, specs
      @ In, mode, string, case mode to operate in (e.g. 'sweep' or 'opt')
      @ In, comp_name, string, name of component this Interaction belongs to
      @ Out, None
    """
    self.raiseADebug(' ... loading interaction "{}"'.format(self.tag))
    self._dispatchable = specs.parameterValues['dispatch']
    for item in specs.subparts:
      name = '_' + item.getName()
      if name in ['_capacity', '_minimum']:
        # common reading for valued params
        self._set_valued_param(name, comp_name, item, mode)
        if name == '_capacity':
          self._capacity_var = item.parameterValues.get('resource', None)
        elif item.getName() == 'minimum':
          self._minimum_var = item.parameterValues.get('resource', None)
    # finalize some values
    resources = set(list(self.get_inputs()) + list(self.get_outputs()))
    ## capacity: if "variable" is None and only one resource in interactions, then that must be it
    if self._capacity_var is None:
      if len(resources) == 1:
        self._capacity_var = list(resources)[0]
      else:
        self.raiseAnError(IOError, 'If multiple resources are active, "capacity" requires a "resource" specified!')
    ## minimum: basically the same as capacity, functionally
    if self._minimum and self._minimum_var is None:
      if len(resources) == 1:
        self._minimum_var = list(resources)[0]
      else:
        self.raiseAnError(IOError, 'If multiple resources are active, "minimum" requires a "resource" specified!')

  def _set_valued_param(self, name, comp, spec, mode):
    """
      Sets up use of a ValuedParam for this interaction for the "name" attribute of this class.
      @ In, name, str, name of member of this class
      @ In, comp, str, name of associated component
      @ In, spec, InputParam, input specifications
      @ In, mode, string, case mode to operate in (e.g. 'sweep' or 'opt')
      @ Out, None
    """
    vp = ValuedParamHandler(name)
    signal = vp.read(comp, spec, mode)
    self._signals.update(signal)
    self._crossrefs[name] = vp
    setattr(self, name, vp)

  def get_capacity(self, meta, raw=False):
    """
      Returns the capacity of this interaction.
      Returns an evaluated value unless "raw" is True, then gives ValuedParam
      @ In, meta, dict, additional variables to pass through
      @ In, raw, bool, optional, if True then provide ValuedParam instead of evaluation
      @ Out, evaluated, float or ValuedParam, requested value
      @ Out, meta, dict, additional variable passthrough
    """
    if raw:
      return self._capacity
    meta['request'] = {self._capacity_var: None}
    evaluated, meta = self._capacity.evaluate(meta, target_var=self._capacity_var)
    return evaluated, meta

  def get_capacity_var(self):
    """
      Returns the resource variable that is used to define the capacity limits of this interaction.
      @ In, None
      @ Out, capacity_var, string, name of capacity-limiting resource
    """
    return self._capacity_var

  def set_capacity(self, cap):
    """
      Allows hard-setting the capacity of this interaction.
      This destroys any underlying ValuedParam that was there before.
      @ In, cap, float, capacity value
      @ Out, None
    """
    self._capacity.set_value(float(cap))

  def get_minimum(self, meta, raw=False):
    """
      Returns the minimum level of this interaction.
      Returns an evaluated value unless "raw" is True, then gives ValuedParam
      @ In, meta, dict, additional variables to pass through
      @ In, raw, bool, optional, if True then provide ValuedParam instead of evaluation
      @ Out, evaluated, float or ValuedParam, requested value
      @ Out, meta, dict, additional variable passthrough
    """
    if raw:
      return self._minimum
    meta['request'] = {self._minimum_var: None}
    if self._minimum is None:
      evaluated = {self._capacity_var: 0.0}
    else:
      evaluated, meta = self._minimum.evaluate(meta, target_var=self._minimum_var)
    return evaluated, meta

  def get_sqrt_RTE(self):
    """
      Provide the square root of the round-trip efficiency for this component.
      Note we use the square root due to splitting loss across the input and output.
      @ In, None
      @ Out, RTE, float, round-trip efficiency as a multiplier
    """
    return self._sqrt_rte

  def get_crossrefs(self):
    """
      Getter.
      @ In, None
      @ Out, crossrefs, dict, resource references
    """
    return self._crossrefs

  def set_crossrefs(self, refs):
    """
      Setter.
      @ In, refs, dict, resource cross-reference objects
      @ Out, None
    """
    # connect references to ValuedParams (Placeholder objects)
    for attr, obj in refs.items():
      valued_param = self._crossrefs[attr]
      valued_param.set_object(obj)
    # perform crosscheck that VPs have what they need
    for attr, vp in self.get_crossrefs().items():
      vp.crosscheck(self)

  def get_inputs(self):
    """
      Returns the set of resources that are inputs to this interaction.
      @ In, None
      @ Out, inputs, set, set of inputs
    """
    return set()

  def get_outputs(self):
    """
      Returns the set of resources that are outputs to this interaction.
      @ In, None
      @ Out, outputs, set, set of outputs
    """
    return set()

  def get_resources(self):
    """
      Returns set of resources used by this interaction.
      @ In, None
      @ Out, resources, set, set of resources
    """
    return list(self.get_inputs()) + list(self.get_outputs())

  def get_tracking_vars(self):
    """
      Provides the variables used by this component to track dispatch
      @ In, None
      @ Out, get_tracking_vars, list, variable name list
    """
    return self._tracking_vars

  def is_dispatchable(self):
    """
      Getter. Indicates if this interaction is Fixed, Dependent, or Independent.
      @ In, None
      @ Out, dispatchable, string, one of 'fixed', 'dependent', or 'independent'
    """
    return self._dispatchable

  def is_type(self, typ):
    """
      Checks if this interaction matches the request.
      @ In, typ, string, name to check against
      @ Out, is_type, bool, whether there is a match or not.
    """
    return typ == self.__class__.__name__

  def is_governed(self):
    """
      Determines if this interaction is optimizable or governed by some function.
      @ In, None
      @ Out, is_governed, bool, whether this component is governed.
    """
    # Default option is False; specifics by interaction type
    return False

  def produce(self, *args, **kwargs):
    """
      Determines the results of this interaction producing resources.
      @ In, args, list, positional arguments
      @ In, kwargs, dict, keyword arguments
      @ Out, None
    """
    raise NotImplementedError('This interaction has no "produce" method.')

  def produce_max(self, *args, **kwargs):
    """
      Determines the results of this interaction producing maximum resources.
      @ In, args, list, positional arguments
      @ In, kwargs, dict, keyword arguments
      @ Out, None
    """
    raise NotImplementedError('This interaction has no produce_max method yet!')

  def produce_min(self, *args, **kwargs):
    """
      Determines the results of this interaction producing minimum resources.
      @ In, args, list, positional arguments
      @ In, kwargs, dict, keyword arguments
      @ Out, None
    """
    raise NotImplementedError('This interaction has no produce_min method yet!')

  def check_expected_present(self, data, expected, premessage):
    """
      checks dict to make sure members are present and not None
      @ In, data, dict, variable set to check against
      @ In, expected, list, list of expected entries
      @ In, premessage, str, prepend message to add to print
      @ Out, None
    """
    # check missing
    missing = list(d for d in expected if d not in data)
    if missing:
      self.raiseAWarning(premessage, '| Expected variables are missing:', missing)
    # check None
    nones = list(d for d, v in data.items() if (v is None and v in expected))
    if nones:
      self.raiseAWarning(premessage, '| Expected variables are None:', nones)
    if missing or nones:
      self.raiseAnError(RuntimeError, 'Some variables were missing or None! See warning messages above for details!')

  def _check_capacity_limit(self, res, amt, balance, meta, raven_vars, dispatch, t):
    """
      Check to see if capacity limits of this component have been violated.
      @ In, res, str, name of capacity-limiting resource
      @ In, amt, float, requested amount of resource used in interaction
      @ In, balance, dict, results of requested interaction
      @ In, meta, dict, additional variable passthrough
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ Out, balance, dict, new results of requested action, possibly modified if capacity hit
      @ Out, meta, dict, additional variable passthrough
    """
    cap = self.get_capacity(meta)[0][self._capacity_var]
    try:
      if abs(balance[self._capacity_var]) > abs(cap):
        #ttttt
        # do the inverse problem: how much can we make?
        balance, meta = self.produce_max(meta, raven_vars, dispatch, t)
        print('The full requested amount ({res}: {req}) was not possible, so accessing maximum available instead ({res}: {blc}).'.format(res=res, req=amt, blc=balance[res]))
    except KeyError:
      raise SyntaxError('Resource "{}" is listed as capacity limiter, but not an output of the component! Got: {}'.format(self._capacity_var, balance))
    return balance, meta

  def get_transfer(self):
    """
      Returns the transfer function, if any
      @ In, None
      @ Out, transfer, transfer ValuedParam
    """
    return self._transfer



class Producer(Interaction):
  """
    Explains a particular interaction, where resources are consumed to produce other resources
  """
  tag = 'produces' # node name in input file

  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, input_specs, InputData, specs
    """
    specs = super().get_input_specs()
    specs.addSub(InputData.parameterInputFactory('consumes', contentType=InputTypes.StringListType, descr=r"""The producer can either produce or consume a resource. If the producer is a consumer it must be accompnied with a transfer function to convert one source of energy to another. """))
    descr = r"""describes how input resources yield output resources for this component's transfer function."""
    specs.addSub(vp_factory.make_input_specs('transfer', descr=descr, kind='transfer'))
    return specs

  def __init__(self, **kwargs):
    """
      Constructor
      @ In, None
      @ Out, None
    """
    Interaction.__init__(self, **kwargs)
    self._produces = []   # the resource(s) produced by this interaction
    self._consumes = []   # the resource(s) consumed by this interaction
    self._tracking_vars = ['production']

  def read_input(self, specs, mode, comp_name):
    """
      Sets settings from input file
      @ In, specs, InputData, specs
      @ In, mode, string, case mode to operate in (e.g. 'sweep' or 'opt')
      @ In, comp_name, string, name of component this Interaction belongs to
      @ Out, None
    """
    # specs were already checked in Component
    Interaction.read_input(self, specs, mode, comp_name)
    self._produces = specs.parameterValues['resource']
    for item in specs.subparts:
      if item.getName() == 'consumes':
        self._consumes = item.value
      elif item.getName() == 'transfer':
        self._set_valued_param('_transfer', comp_name, item, mode)

    # input checking
    ## if a transfer function not given, can't be consuming a resource
    if self._transfer is None:
      if self._consumes:
        self.raiseAnError(IOError, 'Any component that consumes a resource must have a transfer function describing the production process!')

  def get_inputs(self):
    """
      Returns the set of resources that are inputs to this interaction.
      @ In, None
      @ Out, inputs, set, set of inputs
    """
    inputs = Interaction.get_inputs(self)
    inputs.update(np.atleast_1d(self._consumes))
    return inputs

  def get_outputs(self):
    """
      Returns the set of resources that are outputs to this interaction.
      @ In, None
      @ Out, outputs, set, set of outputs
    """
    outputs = set(np.atleast_1d(self._produces))
    return outputs

  def print_me(self, tabs=0, tab='  '):
    """
      Prints info about self
      @ In, tabs, int, optional, number of tabs to insert before prints
      @ In, tab, str, optional, characters to use to denote hierarchy
      @ Out, None
    """
    pre = tab*tabs
    self.raiseADebug(pre+'Producer:')
    self.raiseADebug(pre+'  produces:', self._produces)
    self.raiseADebug(pre+'  consumes:', self._consumes)
    self.raiseADebug(pre+'  transfer:', self._transfer)
    self.raiseADebug(pre+'  capacity:', self._capacity)

  def produce(self, request, meta, raven_vars, dispatch, t, level=None):
    """
      Determines the results of this interaction producing resources.
      @ In, request, dict, requested action {resource: amount}
      @ In, meta, dict, additional variables to pass through
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ In, level, float, storage leven (unused for this Interaction)
      @ Out, balance, dict, results of requested action
      @ Out, meta, dict, additional variable passthrough
    """
    # is there a transfer function to apply?
    res, amt = next(iter(request.items()))
    if self._transfer:
      balance, meta = self.transfer(request, meta, raven_vars, dispatch, t)
    # TODO catch errors! Custom user-based errors?
    else:
      # no transfer function, then we provide the request as-is
      balance = request
    # check if capacity was exceeded
    balance, meta = self._check_capacity_limit(res, amt, balance, meta, raven_vars, dispatch, t)
    return balance, meta

  def produce_max(self, meta, raven_vars, dispatch, t):
    """
      Determines the results of this interaction producing maximum resources.
      @ In, meta, dict, additional variables to pass through
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ Out, balance, dict, results of requested action
      @ Out, meta, dict, additional variable passthrough
    """
    request, meta = self.get_capacity(meta, raven_vars, dispatch, t)
    balance, meta = self.produce(request, meta, raven_vars, dispatch, t)
    # dict((prod, self._capacity[p]) for p, prod in enumerate(self._produces))
    return balance, meta

  def produce_min(self, meta, raven_vars, dispatch, t):
    """
      Determines the results of this interaction producing minimum resources.
      @ In, meta, dict, additional variables to pass through
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ Out, balance, dict, results of requested action
      @ Out, meta, dict, additional variable passthrough
    """
    if self._minimum:
      request, meta = self.get_minimum(meta, raven_vars, dispatch, t)#[self._minimum]
      request = {self._minimum_var: request[self._minimum_var]}
    else:
      request = {next(iter(self.get_outputs())): 0.0}
    balance, meta = self.produce(request, meta, raven_vars, dispatch, t)
    return balance, meta

  def transfer(self, request, meta, raven_vars, dispatch, t):
    """
      Use the transfer function to make a balance of activities that should occur
      @ In, request, dict, requested action {resource: amount}
      @ In, meta, dict, additional variables to pass through
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ Out, balance, dict, results of requested action
      @ Out, meta, dict, additional variable passthrough
    """
    assert len(request) == 1
    balance = defaultdict(float)
    # in the rare case that the transfer function is simple ...
    resources_in = list(self.get_inputs())
    resources_out = list(self.get_outputs())
    inputs = meta
    inputs['request'] = request
    inputs['t'] = t
    inputs['dispatch'] = dispatch
    balance, meta = self._transfer.evaluate(inputs)
    self.check_expected_present(balance, self.get_resources(), 'TRANSFER FUNCTION {}'.format(self._transfer))
    # OLD if transfer evaluation is a float (float, arma), then it signifies a conversion rate
    ## note that we've checked in the input reading for this singular relationship

    if False: #len(balance) == 1:
      requested, rate = list(balance.items())[0] # requested resource and the transfer rate (amount of product per consumed)
      amount = list(requested.values())[0]       # amount of requested resource
      if requested in resources_in:
        balance[resources_out[0]] = -1.0 * rate * amount # NOTE: amount should be negative, but output should be positive
      else:
        balance[inputs[0]] = -1.0 / rate * amount  # NOTE: amount should be positive, but input should be negative
    # check that all values got filled -> TODO remove this for opt performance
    missing = set(resources_in + resources_out) - set(balance.keys())
    if missing:
      self.raiseAnError(RuntimeError, 'While evaluating transfer function, not all variables requested were provided!' +\
                        '  Missing: {}'.format(missing) +\
                        '  Transfer function: {}'.format(self._transfer))
    return balance, meta



class Storage(Interaction):
  """
    Explains a particular interaction, where a resource is stored and released later
  """
  tag = 'stores' # node name in input file

  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, input_specs, InputData, specs
    """
    specs = super().get_input_specs()
    # TODO unused, please implement ... :
    # descr = r"""the limiting charge/discharge rate of this storage. """
    # specs.addSub(ValuedParam.get_input_specs('rate'))
    descr=r"""initial quantity of resource assumed to be present in the storage unit at the beginning
          of a given calculation, in units of quantity (not rate). \default{0}. """
    specs.addSub(vp_factory.make_input_specs('initial_stored', descr=descr))
    descr=r"""control strategy for operating the storage. If not specified, uses a perfect foresight strategy. """
    specs.addSub(vp_factory.make_input_specs('strategy', allowed=['Function'], descr=descr))
    descr = r"""round-trip efficiency for this component as a scalar multiplier. \default{1.0}"""
    specs.addSub(InputData.parameterInputFactory('RTE', contentType=InputTypes.FloatType, descr=descr))
    return specs

  def __init__(self, **kwargs):
    """
      Constructor
      @ In, kwargs, dict, passthrough args
      @ Out, None
    """
    Interaction.__init__(self, **kwargs)
    self._stores = None         # the resource stored by this interaction
    self._rate = None           # the rate at which this component can store up or discharge
    self._initial_stored = None # how much resource does this component start with stored?
    self._strategy = None       # how to operate storage unit
    self._tracking_vars = ['level', 'charge', 'discharge'] # stored quantity, charge activity, discharge activity

  def read_input(self, specs, mode, comp_name):
    """
      Sets settings from input file
      @ In, specs, InputData, specs
      @ In, mode, string, case mode to operate in (e.g. 'sweep' or 'opt')
      @ In, comp_name, string, name of component this Interaction belongs to
      @ Out, None
    """
    # specs were already checked in Component
    Interaction.read_input(self, specs, mode, comp_name)
    self._stores = specs.parameterValues['resource']
    for item in specs.subparts:
      if item.getName() == 'rate':
        self._set_valued_param('_rate', comp_name, item, mode)
      elif item.getName() == 'initial_stored':
        self._set_valued_param('_initial_stored', comp_name, item, mode)
      elif item.getName() == 'strategy':
        self._set_valued_param('_strategy', comp_name, item, mode)
      elif item.getName() == 'RTE':
        self._sqrt_rte = np.sqrt(item.value)
    assert len(self._stores) == 1, 'Multiple storage resources given for component "{}"'.format(comp_name)
    self._stores = self._stores[0]
    # checks and defaults
    if self._initial_stored is None:
      self.raiseAWarning('Initial storage level for "{}" was not provided! Defaulting to 0.'.format(comp_name))
      # make a fake reader node for a 0 value
      vp = ValuedParamHandler('initial_stored')
      vp.set_const_VP(0.0)
      self._initial_stored = vp
    # the capacity is limited by the stored resource.
    self._capacity_var = self._stores

  def get_inputs(self):
    """
      Returns the set of resources that are inputs to this interaction.
      @ In, None
      @ Out, inputs, set, set of inputs
    """
    inputs = Interaction.get_inputs(self)
    inputs.update(np.atleast_1d(self._stores))
    return inputs

  def get_outputs(self):
    """
      Returns the set of resources that are outputs to this interaction.
      @ In, None
      @ Out, outputs, set, set of outputs
    """
    outputs = Interaction.get_outputs(self)
    outputs.update(np.atleast_1d(self._stores))
    return outputs

  def get_resource(self):
    """
      Returns the resource this unit stores.
      @ In, None
      @ Out, stores, str, resource stored
    """
    return self._stores

  def get_strategy(self):
    """
      Returns the resource this unit stores.
      @ In, None
      @ Out, stores, str, resource stored
    """
    return self._strategy

  def is_governed(self):
    """
      Determines if this interaction is optimizable or governed by some function.
      @ In, None
      @ Out, is_governed, bool, whether this component is governed.
    """
    return self._strategy is not None

  def print_me(self, tabs=0, tab='  '):
    """
      Prints info about self
      @ In, tabs, int, optional, number of tabs to insert before prints
      @ In, tab, str, optional, characters to use to denote hierarchy
      @ Out, None
    """
    pre = tab*tabs
    self.raiseADebug(pre+'Storage:')
    self.raiseADebug(pre+'  stores:', self._stores)
    self.raiseADebug(pre+'  rate:', self._rate)
    self.raiseADebug(pre+'  capacity:', self._capacity)

  def produce(self, request, meta, raven_vars, dispatch, t, level=None):
    """
      Determines the results of this interaction producing resources.
      @ In, request, dict, requested action {resource: amount}
      @ In, meta, dict, additional variables to pass through
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ In, level, float, storage level
      @ Out, balance, dict, results of requested action
      @ Out, meta, dict, additional variable passthrough
    """
    if level is None:
      raise RuntimeError('Storage level information was not provided to Storage produce call!')
    dt = dispatch()['time'].values
    dt = dt[t] - dt[t-1] if t > 0 else dt[t+1] - dt[t]
    res, amt_rate = next(iter(request.items()))
    # UNITS: amt_rate is in resource per time, NOT pure resource!
    amt_amount = amt_rate * dt
    assert res == self.get_resource(), 'Requested var is not the same as stored var!'
    balance, meta = self._check_capacity_limit(res, amt_amount, {}, meta, raven_vars, dispatch, t, level)
    # also check rate limit
    delta_amount = balance[res]
    delta_rate = delta_amount / dt # flip the sign to show we're doing as we were asked ...?
    #print('DEBUGG effective battery cons/prod rate:', delta_rate)
    balance, meta = self._check_rate_limit(res, delta_rate, {}, meta, raven_vars, dispatch, t)
    return balance, meta

  def _check_capacity_limit(self, res, amt, balance, meta, raven_vars, dispatch, t, level):
    """
      Check to see if capacity limits of this component have been violated.
      overloads Interaction method, since units for storage are "res" not "res per second"
      @ In, res, str, name of capacity-limiting resource
      @ In, amt, float, requested amount of resource used in interaction
      @ In, balance, dict, results of requested interaction
      @ In, meta, dict, additional variable passthrough
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ In, level, float, current level of storage
      @ Out, balance, dict, new results of requested action, possibly modified if capacity hit
      @ Out, meta, dict, additional variable passthrough
    """
    # note "amt" has units of AMOUNT not RATE (resource, not resource per second)
    sign = np.sign(amt)
    # are we storing or providing?
    #print('DEBUGG supposed current level:', level)
    if sign < 0:
      # we are being asked to consume some
      cap, meta = self.get_capacity(meta, raven_vars, dispatch, t)
      available_amount = cap[res] - level
      #print('Supposed Capacity, Only calculated ins sign<0 (being asked to consumer)',cap)
    else:
      # we are being asked to produce some
      available_amount = level
    # the amount we can consume is the minimum of the requested or what's available
    delta = sign * min(available_amount, abs(amt))
    return {res: delta}, meta

  def _check_rate_limit(self, res, amt, balance, meta, raven_vars, dispatch, t):
    """
      Determines the limiting rate of in/out production for storage
      @ In, res, str, name of capacity-limiting resource
      @ In, amt, float, requested amount of resource used in interaction
      @ In, balance, dict, results of requested interaction
      @ In, meta, dict, additional variable passthrough
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ Out, balance, dict, new results of requested action, possibly modified if capacity hit
      @ Out, meta, dict, additional variable passthrough
    """
    # TODO distinct up/down rates
    # check limiting rate for resource flow in/out, if any
    if self._rate:
      request = {res: None}
      inputs = {'request': request,
                'meta': meta,
                'raven_vars': raven_vars,
                'dispatch': dispatch,
                't': t}
      max_rate = self._rate.evaluate(inputs, target_var=res)[0][res]
      delta = np.sign(amt) * min(max_rate, abs(amt))
      print('max_rate in _check_rate_limit',max_rate, 'delta (min of maxrate and abs(amt)',delta)
      return {res: delta}, meta
    return {res: amt}, meta

  def get_initial_level(self, meta):
    """
      Find initial level of the storage
      @ In, meta, dict, additional variable passthrough
      @ Out, initial, float, initial level
    """
    res = self.get_resource()
    request = {res: None}
    meta['request'] = request
    return self._initial_stored.evaluate(meta, target_var=res)[0][res]



class Demand(Interaction):
  """
    Explains a particular interaction, where a resource is demanded
  """
  tag = 'demands' # node name in input file

  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, input_specs, InputData, specs
    """
    specs = super().get_input_specs()
    return specs

  def __init__(self, **kwargs):
    """
      Constructor
      @ In, kwargs, dict, arguments
      @ Out, None
    """
    Interaction.__init__(self, **kwargs)
    self._demands = None  # the resource demanded by this interaction
    self._penalty = None  # how to penalize for not meeting demand NOT IMPLEMENTED
    self._tracking_vars = ['production']

  def read_input(self, specs, mode, comp_name):
    """
      Sets settings from input file
      @ In, specs, InputData, specs
      @ In, mode, string, case mode to operate in (e.g. 'sweep' or 'opt')
      @ In, comp_name, string, name of component this Interaction belongs to
      @ Out, None
    """
    # specs were already checked in Component
    # must set demands first, so that "capacity" can access it
    self._demands = specs.parameterValues['resource']
    Interaction.read_input(self, specs, mode, comp_name)
    for item in specs.subparts:
      if item.getName() == 'penalty':
        self._set_valued_param('_rate', comp_name, item, mode)

  def get_inputs(self):
    """
      Returns the set of resources that are inputs to this interaction.
      @ In, None
      @ Out, inputs, set, set of inputs
    """
    inputs = Interaction.get_inputs(self)
    inputs.update(np.atleast_1d(self._demands))
    return inputs

  def print_me(self, tabs=0, tab='  '):
    """
      Prints info about self
      @ In, tabs, int, optional, number of tabs to insert before prints
      @ In, tab, str, optional, characters to use to denote hierarchy
      @ Out, None
    """
    pre = tab*tabs
    self.raiseADebug(pre+'Demand/Load:')
    self.raiseADebug(pre+'  demands:', self._demands)
    self.raiseADebug(pre+'  penalty:', self._penalty)
    self.raiseADebug(pre+'  capacity:', self._capacity)

  def produce(self, request, meta, raven_vars, dispatch, t, level=None):
    """
      Determines the results of this interaction producing resources.
      @ In, request, dict, requested action {resource: amount}
      @ In, meta, dict, additional variables to pass through
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ In, level, float, storage level (unused for this Interaction)
      @ Out, balance, dict, results of requested action
      @ Out, meta, dict, additional variable passthrough
    """
    # Q: should this have a transfer function or something? At least capacity limits?
    # A: no; if you want this functionality, add an intervening component with a transfer function.
    res, amt = next(iter(request.items()))
    balance, meta = self._check_capacity_limit(res, amt, request, meta, raven_vars, dispatch, t)
    res, amt = next(iter(balance.items()))
    amt = -1 * abs(amt)
    balance[res] = amt
    return balance, meta

  def produce_max(self, meta, raven_vars, dispatch, t):
    """
      Determines the results of this interaction producing maximum resources.
      @ In, meta, dict, additional variables to pass through
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ Out, balance, dict, results of requested action
      @ Out, meta, dict, additional variable passthrough
    """
    request, meta = self.get_capacity(meta, raven_vars, dispatch, t)
    return request, meta

  def produce_min(self, meta, raven_vars, dispatch, t):
    """
      Determines the results of this interaction producing minimum resources.
      @ In, meta, dict, additional variables to pass through
      @ In, raven_vars, dict, TODO part of meta! consolidate!
      @ In, dispatch, dict, TODO part of meta! consolidate!
      @ In, t, int, TODO part of meta! consolidate!
      @ Out, balance, dict, results of requested action
      @ Out, meta, dict, additional variable passthrough
    """
    if self._minimum:
      request, meta = self.get_minimum(meta, raven_vars, dispatch, t)
    else:
      request = {next(iter(self.get_inputs())): 0.0} # TODO is this a good choice when no min var avail?
    return request, meta



# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Author: dylanjm
  Date: 2021-05-18
"""
import itertools as it
import matplotlib as mpl
mpl.use('Agg') # Prevents the script from blocking while plotting
import matplotlib.pyplot as plt
from typing import List, Dict

try:
  from ravenframework.PluginBaseClasses.OutStreamPlotPlugin import PlotPlugin, InputTypes, InputData
except ModuleNotFoundError:
  import sys
  from . import _utils
  sys.path.append(_utils.get_raven_loc())
  from ravenframework.PluginBaseClasses.OutStreamPlotPlugin import PlotPlugin, InputTypes, InputData


# Matplotlib Global Settings
plt.rc("figure", figsize=(12, 8), titleweight='bold') # type: ignore
plt.rc("axes", titleweight="bold", labelsize=12, axisbelow=True, grid=True) # type: ignore
plt.rc("savefig", bbox="tight") # type: ignore
plt.rc("legend", fontsize=12) # type:ignore
plt.rc(["xtick", "ytick"], labelsize=10) # type: ignore


class DispatchPlot(PlotPlugin):

  @classmethod
  def getInputSpecification(cls):
    """
      Define the acceptable user inputs for this class.
      @ In, None
      @ Out, specs, InputData.ParameterInput,
    """
    specs = super().getInputSpecification()
    specs.addSub(InputData.parameterInputFactory('source', contentType=InputTypes.StringType))
    specs.addSub(InputData.parameterInputFactory('macro_variable', contentType=InputTypes.StringType))
    specs.addSub(InputData.parameterInputFactory('micro_variable', contentType=InputTypes.StringType))
    specs.addSub(InputData.parameterInputFactory('signals', contentType=InputTypes.StringListType))
    return specs

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self.printTag = 'HERON.DispatchPlot'
    self._sourceName = None
    self._source = None
    self._macroName = None
    self._microName = None
    self._addSignals = []

  def handleInput(self, spec):
    """
      Reads in data from the input file
      @ In, spec, InputData.ParameterInput, input information
      @ Out, None
    """
    super().handleInput(spec)
    for node in spec.subparts:
      if node.getName() == 'source':
        self._sourceName = node.value
      elif node.getName() == 'macro_variable':
        self._macroName = node.value
      elif node.getName() == 'micro_variable':
        self._microName = node.value
      elif node.getName() == 'signals':
        self._addSignals = node.value

  def initialize(self, stepEntities):
    """
      Set up plotter for each run
      @ In, stepEntities, dict, entities from the Step
      @ Out, None
    """
    super().initialize(stepEntities)
    src = self.findSource(self._sourceName, stepEntities)
    if src is None:
      self.raiseAnError(IOError, f'Source DataObject "{self._sourceName}" was not found in the Step!')
    self._source = src

  @staticmethod
  def _group_by(iterable: List[str], idx: int) -> Dict[str, List[str]]:
    """
      Return a dictionary containing grouped dispatch variables.
      @ In, iterable, List[str], a list of variable names to group-by.
      @ In, idx, int, the index of the variable to group-by.
      @ Out, gr, Dict[str, List[str]], a dictionary mapping of grouped variable names.
    """
    gr = {}
    for var in iterable:
      # var is expected to have the form: 'Dispatch__component__tracker__resource'
      key = var.split('__')[idx]
      if key in gr.keys():
        gr[key].append(var)
      else:
        gr[key] = [var]
    return gr

  def plot_component(self, fig, axes, df, grp_vars, comp_idx, sid, mstep, cid) -> None:
    """
      Plot and output the optimized dispatch for a specific sample, year, and cluster.
      @ In, fig, matplotlib.figure.Figure, current figure used for plotting.
      @ In, axes, List[List[matplotlib.Axes]], a list of axes to plot each variable.
      @ In, df, pandas.DataFrame, a dataframe containing data to plot.
      @ In, grp_vars, Dict[str, List[str]], a dictionary mapping components to variables.
      @ In, comp_idx, Dict[str, int], a dictionary mapping components to numbers.
      @ In, sid, int, the sample ID.
      @ In, mstep, int, the macro step.
      @ In, cid, int, the cluster ID.
      @ Out, None
    """
    for (key, group), ax in zip(grp_vars.items(), axes.flat):
      lines = []
      for var in group:
        _, comp_name, tracker, _ = var.split('__')
        comp_label = comp_name.replace('_', ' ').title()
        cidx = comp_idx[comp_name]

        # NOTE custom behavior based on production/storage labels
        plot_ax = ax
        var_label = f'{comp_label}, {tracker.title()}'
        ls = '-'
        mk = '1'
        if tracker == 'production':
          var_label = comp_label
        elif tracker == 'level':
          plot_ax = ax.twinx()
          ls = ':'
          mk = '.'
        elif tracker == 'charge':
          mk = '^'
        elif tracker == 'discharge':
          mk = 'v'

        # Plot the micro-step variable on the x-axis (i.e Time)
        ln = plot_ax.plot(df[self._microName], df[var], marker=mk, linestyle=ls, label=var_label, color=f"C{cidx}")
        lines.extend(ln)
        ax.set_title(key.title())
        ax.set_xlabel(self._microName)
        ax.legend(lines, [l.get_label() for l in lines], loc='center left', bbox_to_anchor=(1.03, 0.5))

    # Output and save the image
    file_name = f"dispatch_id{sid}_y{mstep}_c{cid}.png"
    fig.tight_layout()
    fig.savefig(file_name)
    self.raiseAMessage(f'Saved figure to "{file_name}"')
    plt.clf()

  def plot_signal(self, fig, axes, df, sid, mstep, cid) -> None:
    """
      Plot and output the synthetic history for a specific sample, year, and cluster.
      @ In, fig, matplotlib.figure.Figure, a current figure used for plotting.
      @ In, axes, List[List[matplotlib.Axes]], a list of axes to plot each variable.
      @ In, df, pandas.DataFrame, a dataframe containing data to plot.
      @ In, sid, int, the sample ID.
      @ In, mstep, int, the macro step.
      @ In, cid, int, the cluster ID.
    """
    for name, ax in zip(self._addSignals, axes.flat):
      var = df.get(name, None)
      if var is None:
        self.raiseAnError(f'Requested signal variable "{name}" but variable not in data!')
      ax.plot(df[self._microName], var, marker='.', linestyle='-', label=name)
      ax.set_title(name.title())
      ax.set_xlabel(self._microName)
      ax.legend(loc='center left', bbox_to_anchor=(1.03, 0.5))

    signal_file_name = f"dispatch_id{sid}_y{mstep}_c{cid}_SIGNAL.png"
    fig.tight_layout()
    fig.savefig(signal_file_name)
    self.raiseAMessage(f'Saved figure to "{signal_file_name}"')
    plt.clf()

  def run(self):
    """
      Generate the plot
      @ In, None
      @ Out, None
    """
    assert self._source is not None
    ds = self._source.asDataset()
    if ds is None:
      self.raiseAWarning(f'No data in "{self._source.name}" data object; nothing to plot!')
      return
    df = ds.to_dataframe().reset_index()
    dispatch_vars = list(filter(lambda x: "Dispatch__" in x, df.columns))
    grouped_vars = self._group_by(dispatch_vars, -1)
    grouped_comp = self._group_by(dispatch_vars, 1)
    comp_idx = {comp: i for i, comp in enumerate(grouped_comp.keys())}

    # Dimension variables to plot
    sample_ids = df[self._source.sampleTag].unique()
    cluster_ids = df['_ROM_Cluster'].unique()  # TODO: find way to not hardcode name
    macro_steps = df[self._macroName].unique()

    for sample_id, macro_step, cluster_id in it.product(sample_ids, macro_steps, cluster_ids):
      # Filter data to plot correct values for current dimension
      dat = df[
        (df[self._source.sampleTag] == sample_id) &
        (df[self._macroName] == macro_step) &
        (df['_ROM_Cluster'] == cluster_id)
      ]

      # TODO: find a way to combine both plots into one output.
      # Currently, this is difficult to do because of the nested
      # nature of the subplots, as well as the dynamic number of
      # components and signals to plot (i.e. dynamically nested subplots)

      # Output optimized component dispatch for current dimension.
      fig0, axs0 = plt.subplots(len(grouped_vars), 1, sharex=True, squeeze=False)
      self.plot_component(fig0, axs0, dat, grouped_vars, comp_idx, sample_id, macro_step, cluster_id)

      # Output synthetic time series signal for current dimension.
      fig1, axs1 = plt.subplots(len(self._addSignals), 1, sharex=True, squeeze=False)
      self.plot_signal(fig1, axs1, dat, sample_id, macro_step, cluster_id)




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Base class for dispatchers.
"""
from ravenframework.utils import InputData, InputTypes
from ravenframework.BaseClasses import MessageUser, InputDataUser

class Dispatcher(MessageUser, InputDataUser):
  """
    Base class for strategies for consecutive dispatching in a continuous period.
  """
  # ---------------------------------------------
  # INITIALIZATION
  @classmethod
  def get_input_specs(cls):
    """
      Set acceptable input specifications.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = InputData.parameterInputFactory('Dispatcher', ordered=False, baseNode=None)
    # only place specifications that are true for ALL dispatchers here
    return specs

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self.name = 'BaseDispatcher'
    self._time_discretization = None # (start, end, num_steps) to build time discretization
    self._validator = None           # can be used to validate activity

  def read_input(self, inputs):
    """
      Loads settings based on provided inputs
      @ In, inputs, InputData.InputSpecs, input specifications
      @ Out, None
    """
    pass # add here if true for ALL dispatchers

  def initialize(self, case, components, sources, **kwargs):
    """
      Initialize dispatcher properties.
      @ In, case, Case, HERON case instance
      @ In, components, list, HERON components
      @ In, sources, list, HERON sources
      @ In, kwargs, dict, keyword arguments
      @ Out, None
    """
    pass

  # ---------------------------------------------
  # GETTER AND SETTERS
  def get_time_discr(self):
    """
      Retrieves the time discretization information.
      @ In, None
      @ Out, info, tuple, (start, end, number of steps) for time discretization
    """
    return self._time_discretization

  def set_time_discr(self, info):
    """
      Stores the time discretization information.
      @ In, info, tuple, (start, end, number of steps) for time discretization
      @ Out, None
    """
    assert info is not None
    # TODO is this the right idea?
    # don't expand into linspace right now, just store the pieces
    self._time_discretization = info

  def set_validator(self, validator):
    """
      Sets the dispatch validation instance to use in dispatching.
      @ In, validator, HERON Validator, instance of validator
      @ Out, None
    """
    self._validator = validator

  # ---------------------------------------------
  # API
  # TODO make this a virtual method?
  def dispatch(self, case, components, sources):
    """
      Performs technoeconomic dispatch.
      @ In, case, Case, HERON case
      @ In, components, list, HERON components
      @ In, sources, list, HERON sources
      @ Out, results, dict, economic and production metrics
    """
    raise NotImplementedError # must be implemented by inheriting classes

  def validate(self, components, activity, times, meta):
    """
      Method to validate a dispatch activity.
      @ In, components, list, HERON components whose cashflows should be evaluated
      @ In, activity, DispatchState instance, activity by component/resources/time
      @ In, times, np.array(float), time values to evaluate; may be length 1 or longer
      @ In, meta, dict, extra information needed for validation
      @ Out, validation, dict, information about validation
    """
    # default implementation
    if self._validator is not None:
      return self._validator.validate(components, activity, times, meta)
    else:
      # no validator, nothing needs to be changed
      return {}

  # ---------------------------------------------
  # UTILITY METHODS
  def _compute_cashflows(self, components, activity, times, meta, state_args=None, time_offset=0):
    """
      Method to compute CashFlow evaluations given components and their activity.
      @ In, components, list, HERON components whose cashflows should be evaluated
      @ In, activity, DispatchState instance, activity by component/resources/time
      @ In, times, np.array(float), time values to evaluate; may be length 1 or longer
      @ In, meta, dict, additional info to be passed through to functional evaluations
      @ In, state_args, dict, optional, additional arguments to pass while getting activity state
      @ In, time_offset, int, optional, increase time index tracker by this value if provided
      @ Out, total, float, total cashflows for given components
    """
    if state_args is None:
      state_args = {}
    total = 0
    specific_meta = dict(meta) # TODO what level of copying do we need here?
    resource_indexer = meta['HERON']['resource_indexer']
    #print('DEBUGG computing cashflows!')
    for comp in components:
      #print(f'DEBUGG ... comp {comp.name}')
      specific_meta['HERON']['component'] = comp
      comp_subtotal = 0
      for t, time in enumerate(times):
        #print(f'DEBUGG ... ... time {t}')
        # NOTE care here to assure that pyomo-indexed variables work here too
        specific_activity = {}
        for tracker in comp.get_tracking_vars():
          specific_activity[tracker] = {}
          for resource in resource_indexer[comp]:
            specific_activity[tracker][resource] = activity.get_activity(comp, tracker, resource, time, **state_args)
        specific_meta['HERON']['time_index'] = t + time_offset
        specific_meta['HERON']['time_value'] = time
        cfs = comp.get_state_cost(specific_activity, specific_meta, marginal=True)
        time_subtotal = sum(cfs.values())
        comp_subtotal += time_subtotal
      total += comp_subtotal
    return total






# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Interface for user-provided dispatching strategies.
"""
import os
import inspect
import numpy as np

from ravenframework.utils import utils, InputData, InputTypes
from .Dispatcher import Dispatcher
from .DispatchState import NumpyState

class Custom(Dispatcher):
  """
    Base class for strategies for consecutive dispatching in a continuous period.
  """
  # ---------------------------------------------
  # INITIALIZATION
  @classmethod
  def get_input_specs(cls):
    """
      Set acceptable input specifications.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = InputData.parameterInputFactory('custom', ordered=False, baseNode=None)
    specs.addSub(InputData.parameterInputFactory('location', contentType=InputTypes.StringType,
        descr=r"""The hard drive location of the custom dispatcher. Relative paths are taken with
              respect to the HERON run location. Custom dispatchers must implement
              a \texttt{dispatch} method that accepts the HERON case, components, and sources; this
              method must return the activity for each resource of each component."""))
    return specs

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    Dispatcher.__init__(self)
    self.name = 'CustomDispatcher'
    self._usr_loc = None # user-provided path to custom dispatcher module
    self._file = None    # resolved pathlib.Path to the custom dispatcher module

  def read_input(self, inputs):
    """
      Loads settings based on provided inputs
      @ In, inputs, InputData.InputSpecs, input specifications
      @ Out, None
    """
    usr_loc = inputs.findFirst('location')
    if usr_loc is None:
      raise RuntimeError('No <location> provided for <custom> dispatch strategy in <Case>!')
    # assure python extension, omitting it is a user convenience
    if not usr_loc.value.endswith('.py'):
      usr_loc.value += '.py'
    self._usr_loc = os.path.abspath(os.path.expanduser(usr_loc.value))

  def initialize(self, case, components, sources, **kwargs):
    """
      Initialize dispatcher properties.
      @ In, case, Case, HERON case instance
      @ In, components, list, HERON components
      @ In, sources, list, HERON sources
      @ In, kwargs, dict, keyword arguments
      @ Out, None
    """
    start_loc = case.run_dir
    file_loc = os.path.abspath(os.path.join(start_loc, self._usr_loc))
    # check that it exists
    if not os.path.isfile(file_loc):
      raise IOError(f'Custom dispatcher not found at "{file_loc}"! (input dir "{start_loc}", provided path "{self._usr_loc}"')
    self._file = file_loc
    print(f'Loading custom dispatch at "{self._file}"')
    # load user module
    load_string, _ = utils.identifyIfExternalModelExists(self, self._file, '')
    module = utils.importFromPath(load_string, True)
    # check it works as intended
    if not 'dispatch' in dir(module):
      raise IOError(f'Custom Dispatch at "{self._file}" does not have a "dispatch" method!')
    # TODO other checks?

  def dispatch(self, case, components, sources, meta):
    """
      Performs technoeconomic dispatch.
      @ In, case, Case, HERON case
      @ In, components, list, HERON components
      @ In, sources, list, HERON sources
      @ Out, results, dict, economic and production metrics
    """
    # load up time indices
    t_start, t_end, t_num = self.get_time_discr()
    time = np.linspace(t_start, t_end, t_num) # Note we don't care about segment/cluster here
    # load user module
    load_string, _ = utils.identifyIfExternalModelExists(self, self._file, '')
    module = utils.importFromPath(load_string, True)
    state = NumpyState()
    indexer = meta['HERON']['resource_indexer']
    state.initialize(components, indexer, time)
    # run dispatch
    results = module.dispatch(meta, state)
    # TODO: Check to make sure user has uploaded all activity data.
    return state






# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Evaluated signal values for use in HERON
"""
import os
import sys
import abc
import copy

import HERON.src._utils as hutils
from HERON.src.base import Base

try:
  import ravenframework
except ModuleNotFoundError:
  FRAMEWORK_PATH = hutils.get_raven_loc()
  sys.path.append(FRAMEWORK_PATH)
from ravenframework.utils import InputData, InputTypes, utils, xmlUtils

from ravenframework.ROMExternal import ROMLoader


class Placeholder(Base):
  """
    Objects that hold a place in the HERON workflow
    but don't hold value until converted into the RAVEN workflow.
  """
  def __init__(self, **kwargs):
    """
      Constructor.
      @ In, kwargs, dict, passthrough args
      @ Out, None
    """
    Base.__init__(self, **kwargs)
    self.name = None         # identifier
    self._source = None      # name of file? the signal should come from
    self._var_names = None   # LIST of names of output variable from CSV or ARMA
    self._type = None        # source type, such as CSV, pickle, function ...
    self._target_file = None # source file to take data from
    self._workingDir = kwargs['loc'] # where is the HERON input file?

  @classmethod
  @abc.abstractmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, specs, InputData, specs
    """
    pass

  def read_input(self, xml):
    """
      Sets settings from input file
      @ In, xml, xml.etree.ElementTree.Element, input from user
      @ Out, None
    """
    specs = self.get_input_specs()()
    specs.parseNode(xml)
    self.name = specs.parameterValues['name']
    self._source = specs.value
    # check source exists
    if self._source.startswith('%HERON%'):
      # magic word for "relative to HERON root"
      heron_path = hutils.get_heron_loc()
      self._target_file = os.path.abspath(self._source.replace('%HERON%', heron_path))
    elif self._source.startswith('%FARM%'):
      # magic word for "relative to FARM root"
      farm_path = hutils.get_farm_loc()
      self._target_file = os.path.abspath(self._source.replace('%FARM%', farm_path))
    else:
      # check absolute path
      rel_interp = os.path.abspath(os.path.join(self._workingDir, self._source))
      if os.path.isfile(rel_interp):
        self._target_file = rel_interp
      else:
        # check absolute path
        abs_interp = os.path.abspath(self._source)
        if os.path.isfile(abs_interp):
          self._target_file = abs_interp
        else:
          # let relative path trigger the error
          self._target_file = rel_interp
    # check source
    if not os.path.isfile(self._target_file):
      self.raiseAnError(IOError, f'File not found for <DataGenerator><{self._type}> named "{self.name}".' +
                        f'\nLooked in: "{self._target_file}"' +
                        f'\nGiven location: "{self._source}"')
    return specs

  def checkValid(self, case, components, sources):
    """
      Check validity of placeholder given rest of system
      @ In, case, HERON.Case, case
      @ In, case, list(HERON.Component), components
      @ In, sources, list(HERON.Placeholder), sources
      @ Out, None
    """
    pass # overwrite to check

  def print_me(self, tabs=0, tab='  '):
    """
      Prints info about self
      @ In, tabs, int, number of tabs to prepend
      @ In, tab, str, format for tabs
      @ Out, None
    """
    pre = tab*tabs
    self.raiseADebug(pre+'DataGenerator:')
    self.raiseADebug(pre+'  name:', self.name)
    self.raiseADebug(pre+'  source:', self._source)
    self.raiseADebug(pre+'  variables:', self._var_names)

  def is_type(self, typ):
    """
      Checks for matching type
      @ In, typ, str, type to check against
      @ Out, is_type, bool, True if matching request
    """
    # maybe it's not anything we know about
    if typ not in ['ARMA', 'Function', 'ROM', 'CSV']:
      return False
    return eval('isinstance(self, {})'.format(typ))

  def get_variable(self):
    """
      Returns the variable(s) in use for this placeholder.
      @ In, None
      @ Out, var_names, list, variable names
    """
    return self._var_names





class ARMA(Placeholder):
  """
    Placeholder for signals coming from the ARMA
  """
  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = InputData.parameterInputFactory('ARMA', contentType=InputTypes.StringType, ordered=False, baseNode=None,
        descr=r"""This data source is a source of synthetically-generated histories trained by RAVEN.
              The RAVEN ARMA ROM should be trained and serialized before using it in HERON. The text
              of this node indicates the location of the serialized ROM. This location is usually relative
              with respect to the HERON XML input file; however, a full absolute path can be used,
              or the path can be prepended with ``\%HERON\%'' to be relative to the installation
              directory of HERON.""")
    specs.addParam('name', param_type=InputTypes.StringType, required=True,
        descr=r"""identifier for this data source in HERON and in the HERON input file. """)
    specs.addParam('variable', param_type=InputTypes.StringListType, required=True,
        descr=r"""provides the names of the variables from the synthetic history generators that will
              be used in this analysis.""")
    # TODO someday read this directly off the model instead of asking the user!
    specs.addParam('evalMode', param_type=InputTypes.StringType, required=False,
        descr=r"""desired sampling mode for the ARMA. See the RAVEN manual for options. \default{clustered}""")
    return specs

  def __init__(self, **kwargs):
    """
      Constructor.
      @ In, kwargs, dict, passthrough args
      @ Out, None
    """
    Placeholder.__init__(self, **kwargs)
    self._type = 'ARMA'
    self._var_names = None # variables from the ARMA to use
    self.eval_mode = None # ARMA evaluation style (clustered, full, truncated)
    self.needs_multiyear = None # if not None, then this is a 1-year ARMA that needs multiyearing
    self.limit_interp = None # if not None, gives the years to limit this interpolated ROM to

  def read_input(self, xml):
    """
      Sets settings from input file
      @ In, xml, xml.etree.ElementTree.Element, input from user
      @ Out, None
    """
    specs = Placeholder.read_input(self, xml)
    self._var_names = specs.parameterValues['variable']
    self.eval_mode = specs.parameterValues.get('evalMode', 'clustered')
    # check that the source ARMA exists

  def checkValid(self, case, components, sources):
    """
      Check validity of placeholder given rest of system
      @ In, case, HERON.Case, case
      @ In, case, list(HERON.Component), components
      @ In, sources, list(HERON.Placeholder), sources
      @ Out, None
    """
    self.raiseAMessage(f'Checking ROM at "{self._target_file}"')
    structure = hutils.get_synthhist_structure(self._target_file)
    interpolated = 'macro' in structure
    clustered = bool(structure['clusters'])
    # segmented = bool(structure['segments']) # TODO
    self.raiseAMessage(
        f'For DataGenerator <{self._type}> "{self.name}", detected: {"" if interpolated else "NOT"} interpolated, ' +
        f'{"" if clustered else "NOT"} clustered.'
    )
    # expect that project life == num macro years
    project_life = hutils.get_project_lifetime(case, components) - 1 # one less for construction year
    if interpolated:
      # if interpolated, needs more checking
      interp_years = structure['macro']['num']
      if interp_years >= project_life:
        self.raiseADebug(
            f'"{self.name}" interpolates {interp_years} macro steps,' +
            f'and project life is {project_life}, so histories will be trunctated.'
        )
        self.limit_interp = project_life
      else:
        self.raiseAnError(
            RuntimeError, f'"{self.name}" interpolates {interp_years} macro steps, but project life is {project_life}!'
        )
    else:
      # if single year, we can use multiyear so np
      self.raiseADebug(
          f'"{self.name}" will be extended to project life ({project_life}) macro steps using <Multicycle>.'
      )
      self.needs_multiyear = project_life




class Function(Placeholder):
  """
    Placeholder for values that are evaluated on the fly
  """
  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = InputData.parameterInputFactory('Function', contentType=InputTypes.StringType,
        ordered=False, baseNode=None,
        descr=r"""This data source is a custom Python function to provide derived values.
              Python functions have access to the variables within the dispatcher. The text
              of this node indicates the location of the python file. This location is usually relative
              with respect to the HERON XML input file; however, a full absolute path can be used,
              or the path can be prepended with ``\%HERON\%'' to be relative to the installation
              directory of HERON.""")
    specs.addParam('name', param_type=InputTypes.StringType, required=True,
        descr=r"""identifier for this data source in HERON and in the HERON input file. """)
    return specs

  def __init__(self, **kwargs):
    """
      Constructor.
      @ In, kwargs, dict, passthrough args
      @ Out, None
    """
    Placeholder.__init__(self, **kwargs)
    self._type = 'Function'
    self._module = None
    self._module_methods = {}

  def __getstate__(self):
    """
      Serialization.
      @ In, None
      @ Out, d, dict, object contents
    """
    # d = super(self, __getstate__) TODO only if super has one ...
    d = copy.deepcopy(dict((k, v) for k, v in self.__dict__.items() if k not in ['_module','_module_methods']))
    return d

  def __setstate__(self, d):
    """
      Deserialization.
      @ In, d, dict, object contents
      @ Out, None
    """
    self.__dict__ = d
    self._module = None
    self._module_methods = {}
    target_dir = os.path.dirname(os.path.abspath(self._target_file))
    if target_dir not in sys.path:
      sys.path.append(target_dir)
    load_string, _ = utils.identifyIfExternalModelExists(self, self._target_file, '')
    module = utils.importFromPath(load_string, True)
    if not module:
      raise IOError(f'Module "{self._source}" for function "{self.name}" was not found!')
    self._module = module
    self._set_callables(module)

  def read_input(self, xml):
    """
      Sets settings from input file
      @ In, xml, xml.etree.ElementTree.Element, input from user
      @ Out, None
    """
    Placeholder.read_input(self, xml)
    # load module
    load_string, _ = utils.identifyIfExternalModelExists(self, self._target_file, '')
    module = utils.importFromPath(load_string, True)
    if not module:
      raise IOError(f'Module "{self._source}" for function "{self.name}" was not found!')
    self._module = module
    # TODO do we need to set the var_names? self._var_names = _var_names
    self._set_callables(module)

  def _set_callables(self, module):
    """
      Build a dict of callable methods with the right format
      @ In, module, python Module, module to load methods from
      @ Out, None
    """
    for name, member in module.__dict__.items():
      # check all conditions for not acceptable formats; if none of those true, then it's a good method
      ## check callable as a function
      if not callable(member):
        continue
      self._module_methods[name] = member

  def evaluate(self, method, request, data_dict):
    """
      Evaluates requested method in stored module.
      @ In, method, str, method name
      @ In, request, dict, requested action
      @ In, data_dict, dict, dictonary of evaluation parameters (metadata)
      @ Out, result, dict, results of evaluation
    """
    result = self._module_methods[method](request, data_dict)
    if not (hasattr(result, '__len__') and len(result) == 2 and all(isinstance(r, dict) for r in result)):
      raise RuntimeError('From Function "{f}" method "{m}" expected {s}.{m} '.format(f=self.name, m=method, s=self._source) +\
                         'to return with form (results_dict, meta_dict) with both as dictionaries, but received:\n' +\
                         '    {}'.format(result))
    return result



class ROM(Placeholder):
  """
    Placeholder for values that are evaluated via a RAVEN ROM
  """
  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = InputData.parameterInputFactory('ROM', contentType=InputTypes.StringType,
        ordered=False, baseNode=None,
        descr=r"""This data source is a trained RAVEN ROM to provide derived values.
              Variables within the dispatcher act as sources for the ROM inputs. The text
              of this node indicates the location of the serialized ROM. This location is usually
              relative with respect to the HERON XML input file; however, a full absolute path can
              be used, or the path can be prepended with ``\%HERON\%'' to be relative to the
              installation directory of HERON.""")
    specs.addParam('name', param_type=InputTypes.StringType, required=True,
        descr=r"""identifier for this data source in HERON and in the HERON input file. """)
    return specs

  def __init__(self, **kwargs):
    """
      Constructor.
      @ In, kwargs, dict, passthrough args
      @ Out, None
    """
    super().__init__(**kwargs)
    self._type = 'ROM'
    self._rom = None          # actual unpickled instance of ROM
    self._rom_location = None # string path to ROM

  def read_input(self, xml):
    """
      Sets settings from input file
      @ In, xml, xml.etree.ElementTree.Element, input from user
      @ Out, None
    """
    super().read_input(xml)

    self._runner = ROMLoader(self._target_file)
    # TODO is this serializable? or get/set state for this?

  def evaluate(self, rlz):
    """
      Evaluates requested method in stored module.
      @ In, rlz, dict, input realization as {input_name: value}
      @ Out, result, dict, results of evaluation
    """
    result = self._runner.evaluate(rlz)[0] # [0] because can batch evaluate, I think
    return result


class CSV(Placeholder):
  """
    Placeholder for values taken from a comma-separated file.
  """

  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, specs, InputData, specs
    """
    specs = InputData.parameterInputFactory(
      "CSV",
      contentType=InputTypes.StringType,
      ordered=False,
      baseNode=None,
      descr="""This data source is a static comma separated values (CSV) file.
               The text of this node indicates the location of the CSV file.
               This location is usually relative with respect to the HERON XML input file;
               however, a full absolute path can be used, or the path can be prepended
               with ``\%HERON\%'' to be relative to the installation directory of HERON.
               It is expected that variables contained in this file are defined as headers
               in the first row."""
    )

    specs.addParam(
      "name",
      param_type=InputTypes.StringType,
      required=True,
      descr="""identifier for this data source in HERON and in the HERON input file.""",
    )

    specs.addParam(
      'variable',
      param_type=InputTypes.StringListType,
      required=True,
      descr="""provides the names of the variables found in the CSV file
               and will be used in the workflow. Please note that all CSV files
               used in HERON for the purpose of input data must contain a variable
               titled ``RAVEN_sample_ID''. This variable can be a column of constant
               values (i.e., 0 or 1). This variable is unlikely to be used in the workflow
               but is required by RAVEN.""",
    )
    return specs

  def __init__(self, **kwargs):
    """
      Constructor
      @ In, kwargs, dict, passthrough args
      @ Out, None
    """
    super().__init__(**kwargs)
    self._type = 'CSV'
    self._data = None
    self.eval_mode = "full"
    self.needs_multiyear = 1
    self.limit_interp = 1

  def read_input(self, xml):
    """
      Sets settings from input file.
      @ In, xml, xml.etree.ElementTree.Element, input from user.
      @ Out, None
    """
    specs = super().read_input(xml)
    self._var_names = specs.parameterValues['variable']
    with open(self._target_file, 'r', encoding='utf-8-sig') as f:
      headers = list(s.strip() for s in f.readline().split(','))
    for var in self._var_names:
      if var not in headers:
        self.raiseAnError(
          KeyError,
          f'Variable "{var}" requested for "{self.name}" but not found in "{self._target_file}"! Found: {headers}'
        )

  def checkValid(self, case, components, sources):
    """
      Check validity of placeholder given rest of system
      @ In, case, HERON.Case, case
      @ In, components, list(HERON.Component), components
      @ In, sources, list(HERON.Placeholder), sources
      @ Out, None
    """
    self.raiseAMessage(f'Checking CSV at "{self._target_file}"')
    structure = hutils.get_csv_structure(self._target_file, case.get_year_name(), case.get_time_name())
    interpolated = 'macro' in structure
    clustered = bool(structure['clusters'])
    # segmented = bool(structure['segments']) # TODO
    self.raiseAMessage(
        f'For DataGenerator <{self._type}> "{self.name}", detected: {"" if interpolated else "NOT"} interpolated, ' +
        f'{"" if clustered else "NOT"} clustered.'
    )
    # expect that project life == num macro years
    project_life = hutils.get_project_lifetime(case, components) - 1 # one less for construction year
    if interpolated:
      # if interpolated, needs more checking
      interp_years = structure['macro']['num']
      if interp_years >= project_life:
        self.raiseADebug(
            f'"{self.name}" interpolates {interp_years} macro steps, and project life is {project_life}, so histories will be trunctated.'
        )
        self.limit_interp = project_life
      else:
        self.raiseAnError(
            RuntimeError, f'"{self.name}" interpolates {interp_years} macro steps, but project life is {project_life}!'
        )
    else:
      # if single year, we can use multiyear so np
      self.raiseADebug(
          f'"{self.name}" will be extended to project life ({project_life}) macro steps using <Multicycle>.'
      )
      self.needs_multiyear = project_life




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Values that are expressed as linear ratios of one another.
  Primarily intended for transfer functions.
"""
from .ValuedParam import ValuedParam, InputData, InputTypes

# class for custom dynamically-evaluated quantities
class Linear(ValuedParam):
  """
    Represents a ValuedParam that is a linearized transfer function.
  """

  @classmethod
  def get_input_specs(cls):
    """
      Template for parameters that can take a scalar, an ARMA history, or a function
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory('linear', contentType=InputTypes.StringType,
        descr=r"""indicates this value should be interpreted as a ratio based on an input value.""")
    rate = InputData.parameterInputFactory('rate', contentType=InputTypes.FloatType,
        descr=r"""linear coefficient for the indicated \xmlAttr{resource}.""")
    rate.addParam('resource', param_type=InputTypes.StringType,
        descr=r"""indicates the resource for which the linear transfer rate is being provided in this node.""")
    spec.addSub(rate)
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._coefficients = None # ratios, stored in a dict as key: value

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read valued param from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    super().read(comp_name, spec, mode, alias_dict=None)
    self._coefficients = {}
    for rate_node in spec.findAll('rate'):
      resource = rate_node.parameterValues['resource']
      self._coefficients[resource] = rate_node.value
    return []

  def get_coefficients(self):
    """
      Returns linear coefficients.
      @ In, None
      @ Out, coeffs, dict, coefficient mapping
    """
    return self._coefficients

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, stuff from RAVEN, particularly including the keys 'meta' and 'raven_vars'
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, balance, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, inputs, dict, dictionary of meta (possibly changed during evaluation)
    """
    if target_var not in self._coefficients:
      self.raiseAnError(RuntimeError, f'"rate" for target variable "{target_var}" not found for ' +
                        f'ValuedParam {self.name}!')
    req_res, req_amt = next(iter(inputs['request'].items()))
    req_rate = self._coefficients[req_res]
    balance = {req_res: req_amt}
    for res, rate in self._coefficients.items():
      if res == req_res:
        continue
      balance[res] = rate / req_rate * req_amt
    return balance, inputs




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
from .pyomo_dispatch import Pyomo
from .CustomDispatcher import Custom

known = {
    'pyomo': Pyomo,
    'custom': Custom,
}

def get_class(typ):
  """
    Returns the requested dispatcher type.
    @ In, typ, str, name of one of the dispatchers
    @ Out, class, object, class object
  """
  return known.get(typ, None)




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the base class ValuedParam entity.
"""
from __future__ import unicode_literals, print_function
import os
import sys

from HERON.src import _utils as hutils
from HERON.src.ValuedParams import Parametric
from HERON.src.ValuedParams import factory as VPFactory

try:
  import ravenframework
except ModuleNotFoundError:
  framework_path = hutils.get_raven_loc()
  sys.path.append(framework_path)
from ravenframework.utils import InputData, InputTypes
from ravenframework.BaseClasses import MessageUser

class ValuedParamHandler(MessageUser):
  """
    This class enables the identification of runtime-evaluated variables
    with a variety of sources (fixed values, parametric values, data histories, function
    evaluations, etc).
  """
  @classmethod
  def get_input_specs(cls, name, descr=""):
    """
      Template for parameters that can take a scalar, an ARMA history, or a function
      @ In, name, string, name for spec (tag)
      @ In, descr, string, base description for item (this will add to it)
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory(name,
        descr=descr + r"""This value can be taken from any \emph{one} of the sources described below.""")
    # VP sources
    for vp_type in VPFactory.knownTypes():
      spec.addSub(vp_type.get_input_specs())
    # addons
    spec.addSub(InputData.parameterInputFactory('multiplier', contentType=InputTypes.FloatType,
        descr=r"""Multiplies any value obtained by this parameter by the given value. \default{1}"""))
    # for when the result obtained needs to grow from year to year
    # TODO
    # growth = InputData.parameterInputFactory('growth', contentType=InputTypes.FloatType,
    #     descr=r"""if this node is given, the value will be adjusted from cycle to cycle by the provided amount.""")
    # growth_mode = InputTypes.makeEnumType('growthType', 'growthType', ['linear', 'exponential'])
    # growth.addParam('mode', param_type=growth_mode, required=True,
    #     descr=r"""determines whether the growth factor should be taken as linear or exponential (compounding).""")
    # spec.addSub(growth)
    return spec

  def __init__(self, name):
    """
      Constructor.
      @ In, name, str, name of this valued param
      @ Out, None
    """
    super().__init__()
    self.name = name         # member whom this ValuedParam provides values, e.g. Component.economics.alpha
    self._vp = None          # ValuedParam instance
    self._multiplier = None  # scalar multiplier for evaluation values
    self._growth_val = None  # used to grow the value year-by-year
    self._growth_mode = None # mode for growth (e.g. exponenetial, linear)

  def __repr__(self):
    """
      String representation of this Handler and its VP
      @ In, None
      @ Out, repr, str, string representation
    """
    msg = f'<ValuedParam name: "{self.name}" type: "{self.type}">'
    return msg

  @property
  def type(self):
    """
      Provides type of underlying vp
      @ In, None
      @ Out, type, str, string type (class name) of VP
    """
    return self._vp.type

  def read(self, comp_name: str, spec: InputData.ParameterInput, mode: str, alias_dict=None):
    """
      Used to read valued param from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, signal, list, signals needed to evaluate this ValuedParam at runtime
    """
    # aliases get used to convert variable names, notably for the cashflow's "capacity"
    if alias_dict is None:
      alias_dict = {}
    # instantiate the requested ValuedParam
    found = False
    knownVPs = VPFactory.knownTypes()
    for sub in spec.subparts:
      # ValuedParam
      if sub.getName() in knownVPs:
        # check against multiple source specifications
        if found:
          self.raiseAnError(IOError, 'Only one ValuedParam type can be used per node; received ' +
                            f'multiple for comp "{comp_name}" node <{spec.getName()}>!')
        self._vp = VPFactory.returnInstance(sub.getName())
        signal = self._vp.read(comp_name, sub, mode, alias_dict=alias_dict)
        found = True
      ## other addons
      # multiplier
      elif sub.getName() == 'multiplier':
        self._multiplier = sub.value
      # growth # TODO this doesn't appear to be implemented currently
      # elif sub.getName() == 'growth':
      #   self._growth_val = sub.value
      #   self._growth_mode = sub.parameterValues['mode']
    if not found:
      self.raiseAnError(IOError, f'Component "{comp_name}" node <{spec.getName()}> expected a ValuedParam ' +
                       f'to define its value source, but none was found! Options include: {knownVPs}')
    return signal

  def crosscheck(self, interaction):
    """
      Perform checks to make sure VP is set up correctly.
      @ In, interaction, Component.Interaction, HERON Interaction that "owns" this VP
      @ Out, None (error raised if not correct)
    """
    self._vp.crosscheck(interaction)

  def get_source(self):
    """
      Access the source type and name of the VP
      @ In, None
      @ Out, kind, str, identifier for the expected DataGenerator type (as a string)
      @ Out, source_name, identifier for the name of the associated DataGenerator
    """
    return self._vp.get_source()

  def get_value(self, **kwargs):
    """
      Get fixed values, if any.
      @ In, None
      @ Out, value, object, VP's reported fixed value (may be None)
    """
    return self._vp.get_value(**kwargs)

  def set_value(self, value):
    """
      Get fixed values, if any.
      @ In, value, object, VP's new fixed value (may be None)
      @ Out,
    """
    # should this be allowed if it's not parametric?
    assert isinstance(self._vp, Parametric)
    return self._vp.set_value(value)

  def get_coefficients(self):
    """
      Provide coefficients for linear transformation
      @ In, None
      @ Out, coeffs, dict, coefficients
    """
    assert self.type == 'Linear'
    return self._vp.get_coefficients()

  def is_parametric(self):
    """
      Tell if VP is parametric type
      @ In, None
      @ Out, is, bool, True if parametric
    """
    return isinstance(self._vp, Parametric)

  def set_const_VP(self, value):
    """
      Force the Handler to set a ValuedParam without doing reading.
      Mostly for testing.
      @ In, value, float, fixed value
      @ Out, None
    """
    self._vp = VPFactory.returnInstance('fixed_value')
    self._vp.set_value(value)

  def set_transfer_VP(self, func):
    """
      Force the Handler to set a transfer ValuedParam without doing reading.
      Mostly for testing.
      @ In, func, function, transfer function
      @ Out, None
    """
    from HERON.src.Placeholders import Function
    self._vp = VPFactory.returnInstance('Function')
    self._vp.evaluate = func

  def set_object(self, obj):
    """
      Provides a reference to the requested object for the VP.
      Requests are made through the return of "self.read".
      @ In, obj, isntance, evaluation target
      @ Out, None
    """
    self._vp.set_object(obj)

  def evaluate(self, *args, util_factor=False, **kwargs):
    """
      Evaluate the ValuedParam, wherever it gets its data from
      @ In, args, list, positional arguments for ValuedParam
      @ In, kwargs, dict, keyword arguements for ValuedParam
      @ Out, evaluate, object, stuff from ValuedParam evaluation
    """
    data, meta = self._vp.evaluate(*args, **kwargs)
    if self._multiplier is not None:
      for key in data:
        data[key] *= self._multiplier
    return data, meta

